# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import os
import sys
import asyncio
import readline
import traceback
import platform
import rlcompleter

from nessaid_cli.elements import EndOfInpuToken
from nessaid_cli.interface import CliInterface, ParsingResult
from nessaid_cli.tokenizer.tokenizer import NessaidCliTokenizer, TokenizerException
from nessaid_cli.tokens import CliToken, MATCH_SUCCESS, MATCH_PARTIAL, MATCH_AMBIGUOUS


try:
    import curses
    CURSES_KEY_BACKSPACE = curses.KEY_BACKSPACE
except:
    CURSES_KEY_BACKSPACE = 263

try:
    import getch
    def getinput(prompt, show_char=False):
        """Replacement for getpass.getpass() which prints asterisks for each character typed"""
        print(prompt, end='', flush=True)

        buf = ''
        kbd_int = False
        while True:
            try:
                try:
                    ch = getch.getch()
                except KeyboardInterrupt:
                    buf = ""
                    kbd_int = True
                    break
                except OverflowError as e:
                    continue
            except KeyboardInterrupt:
                buf = ""
                kbd_int = True
                break
            except OverflowError as e:
                continue

            if ch == '\n':
                print('')
                break
            elif ch in (CURSES_KEY_BACKSPACE, '\b', '\x7f'):
                if len(buf) > 0:
                    buf = buf[:-1]
                    print('\b', end='', flush=True)
                    print(" ", end='', flush=True)
                    print('\b', end='', flush=True)
            else:
                buf += ch
                c = '*' if show_char is False else ch
                print(c, end='', flush=True)
        if kbd_int:
            print("\n", end='', flush=True)
        return buf

except ImportError:
    import getpass as gp

    def getinput(prompt, show_char=True):
        if show_char:
            return input(prompt)

        return gp.getpass(prompt)


class NessaidCli(CliInterface):

    def __init__(self, grammarset, loop=None, prompt=None,
                 stdin=None, stdout=None, stderr=None, completekey='tab', use_rawinput=True):

        super().__init__(loop, grammarset, stdin=stdin, stdout=stdout, stderr=stderr)

        self._match_loop = asyncio.new_event_loop()

        self._completekey = completekey
        self._use_rawinput = use_rawinput
        self._cmdqueue = []
        self._old_completer = None
        self._completion_matches = []
        self._exit_loop = False
        self._exec_inited = False
        self._current_line = None
        self._prompt = "" if not prompt else prompt
        self._empty_line_matching = False
        self._suggestion_shown = False
        self._waiting_input = False

        self._loop = loop if loop else asyncio.get_event_loop()
        self._nessaid_tokenizer = NessaidCliTokenizer()

    @property
    def loop(self):
        return self._loop

    @property
    def prompt(self):
        return self._prompt

    def tokenize(self, line):
        try:
            if line == '\n':
                line = ""

            try:
                tokens = self._nessaid_tokenizer.parse(line)
                return True, None, tokens
            except TokenizerException as e:
                return False, "Invalid input: {}".format(line), []
        except Exception as e:
            return False, "Exception parsing line: {}: {}".format(type(e), e)

    async def get_next_line(self, prompt):
        if self._cmdqueue:
            line = self._cmdqueue.pop(0)
        elif self._use_rawinput:
            try:
                line = await self.loop.run_in_executor(None, input, prompt)
            except EOFError:
                line = ""
        else:

            def _readline():
                self.stdout.write(prompt)
                self.stdout.flush()
                line = self.stdin.readline()
                if not len(line):
                    line = ""
                else:
                    line = line.rstrip('\r\n')
                return line

            try:
                line = await self.loop.run_in_executor(None, _readline)
            except Exception:
                line = ""

        return line

    def complete(self, text, state):

        if self._waiting_input:
            readline.insert_text("\t")
            return None

        TOKEN_SEPARATORS = [" "]
        DEFAULT_SEPARATOR = " "

        if state != 0:
            try:
                if self._completion_matches and len(self._completion_matches) > state:
                    return self._completion_matches[state]
                else:
                    return None
            except IndexError:
                return None

        line = readline.get_line_buffer()

        try:
            success, error, tokens = self.tokenize(line)
            if not success:
                self.set_completion_tokens(["Failure tokenizing input line: {}".format(line)])
                return self._completion_matches[state]
        except Exception as e:
            self.set_completion_tokens(["Exception tokenizing input line: {}: {}".format(type(e), e)])
            return None

        if line and line[-1] in TOKEN_SEPARATORS:
            last_token_complete = True
        else:
            last_token_complete = False

        try:
            input_tokens = [str(t) for t in tokens]

            match_output = self._match_loop.run_until_complete(
                    self.match(input_tokens, dry_run=True, last_token_complete=last_token_complete)
                )

            completions = []

            def cli_startswith(string, substring):
                if string.startswith(substring):
                    return True
                else:
                    if match_output.case_insensitive:
                        if string.lower().startswith(substring.lower()):
                            return True
                return False

            has_full_match = False
            if match_output.next_tokens and EndOfInpuToken() in match_output.next_tokens:
                match_output.next_tokens.remove(EndOfInpuToken())
                has_full_match = True

            if match_output.result == MATCH_SUCCESS:
                completions += ["NEWLINE: Complete command"]
            elif match_output.result in [MATCH_PARTIAL, MATCH_AMBIGUOUS]:
                completions = match_output.next_tokens
                if has_full_match:
                    completions += ["NEWLINE: Complete command"]
                if match_output.last_token:
                    tok_input = match_output.last_token[0]
                    tok_replacement = match_output.last_token[1]

                    if tok_replacement and cli_startswith(tok_replacement, tok_input):
                        if tok_input != tok_replacement:
                            if line == self._current_line:
                                idx = line.rindex(tok_input)
                                replace_len = len(line) - idx
                                readline.insert_text(tok_replacement[replace_len:])
                                self._suggestion_shown = False
                                return None
                            elif self._suggestion_shown and self._current_line and len(line) > len(self._current_line):
                                if cli_startswith(line, self._current_line):
                                    idx = line.rindex(tok_input)
                                    replace_len = len(line) - idx
                                    readline.insert_text(tok_replacement[replace_len:])
                                    if len(completions) == 1 and completions[0] == tok_replacement:
                                        readline.insert_text(DEFAULT_SEPARATOR)
                                    self._suggestion_shown = False
                                    return None
                        elif match_output.next_constant_token and match_output.next_constant_token == tok_input:
                            if line and line[-1] not in TOKEN_SEPARATORS:
                                readline.insert_text(DEFAULT_SEPARATOR)
                                self._suggestion_shown = False
                                return None
                elif match_output.next_constant_token:
                    if line and line[-1] in TOKEN_SEPARATORS:
                        if line == self._current_line:
                            readline.insert_text(match_output.next_constant_token)
                            self._suggestion_shown = False
                            return None
                    elif not line and self._empty_line_matching:
                        self._empty_line_matching = False
                        readline.insert_text(match_output.next_constant_token)
                        self._suggestion_shown = False
                        return None
            else:
                self.set_completion_tokens(["Failed to parse: {}".format(match_output.error)])
                return self._completion_matches[state]

            self.set_completion_tokens(completions)
            if completions and match_output.last_token:
                self._suggestion_shown = True

            if not line:
                self._empty_line_matching = True
            else:
                self._empty_line_matching = False
            self._current_line = line

            return self._completion_matches[state]

        except Exception as e:
            self.set_completion_tokens(["Exception matchinsinput line: {}: {}".format(type(e), e)])
            return self._completion_matches[state]

        return None

    def process_completions(self, status, completions):
        pass

    def set_completion_tokens(self, tokens):
        """
        comp_tokens = []
        char_repl_maps = {
            "\\": "\\\\",
            "\n": "\\n",
            "\t": "\\t",
            '"': '\\"'
        }
        for tok in tokens:
            if tok.startswith('"'):
                comp_token = tok[1:]
                if tok.endswith('"'):
                    comp_token = comp_token[:-1]
                    e = '"'
                else:
                    e = ""
                for r, repl in char_repl_maps.items():
                    comp_token = comp_token.replace(r, repl)
                comp_tokens.append('"' + comp_token + e)
            else:
                comp_tokens.append(tok)
                """
        comp_tokens = tokens
        if len(comp_tokens) > 1:
            if os.name == 'nt':
                comp_tokens = sorted(comp_tokens)
                comp_tokens = [comp_tokens[0]] + ['\n' + t for t in comp_tokens[1:]]
        self._completion_matches = comp_tokens + [" "]

    def input(self, prompt="", show_char=True):

        try:
            self._waiting_input = True
            return getinput(prompt, show_char)
        except Exception:
            return ""
        finally:
            self._waiting_input = False

    async def get_input(self, prompt="", show_char=True):

        try:
            self._waiting_input = True
            return await self.loop.run_in_executor(None, getinput, prompt, show_char)
        except Exception:
            return ""
        finally:
            self._waiting_input = False

    async def preloop(self):
        pass

    def exit_loop(self):
        """Exit thr running Cli loop"""
        self._exit_loop = True

    async def exec_args(self, *args):
        line = " ".join(args)
        return await self.exec_line(line)

    async def exec_line(self, line):
        try:
            success, error, tokens = self.tokenize(line)
            if not success:
                self.error("Failure tokenizing input line:", error)
        except Exception as e:
            self.error("Exception tokenizing input line:", type(e), e)
            self.error("\n")
            traceback.print_tb(e.__traceback__, file=self.stderr)
            self.error("\n")

        try:
            arglist = []
            input_tokens = [str(t) for t in tokens]
            match_output = await self.match(input_tokens, dry_run=False, last_token_complete=True, arglist=arglist)
            self.process_cli_response(tokens, match_output)
        except Exception as e:
            self.error("Exception parsing input line:", type(e), e)
            self.error("\n")
            traceback.print_tb(e.__traceback__, file=self.stderr)
            self.error("\n")

    async def cli_exec_init(self):
        if not self._exec_inited:
            if self._use_rawinput and self._completekey:
                self._old_completer = readline.get_completer()
                readline.set_completer(self.complete)
                if readline.__doc__ and 'libedit' in readline.__doc__:
                    readline.parse_and_bind("bind ^I rl_complete")
                else:
                  readline.parse_and_bind(self._completekey+": complete")
            self._exec_inited = True

    async def cmdloop(self, grammarname, intro=None):
        self.enter_grammar(grammarname)
        try:
            await self.preloop()
            await self.cli_exec_init()
            if intro is not None:
                self.stdout.write(str(intro)+"\n")
                self.stdout.flush()

            while not self._exit_loop:
                line = await self.get_next_line(self.prompt)
                self._current_line = line
                #print("Input:", line)
                try:
                    success, error, tokens = self.tokenize(line)
                    if not success:
                        self.error("Failure tokenizing input line:", error)
                except Exception as e:
                    self.error("Exception tokenizing input line:", type(e), e)
                    self.error("\n")
                    traceback.print_tb(e.__traceback__, file=self.stderr)
                    self.error("\n")
                else:
                    #print("Tokens:", tokens)
                    pass

                try:
                    arglist = []
                    input_tokens = [str(t) for t in tokens]
                    match_output = await self.match(input_tokens, dry_run=False, last_token_complete=True, arglist=arglist)
                    self._current_line = None
                    self.process_cli_response(tokens, match_output)
                except Exception as e:
                    self.error("Exception parsing input line:", type(e), e)
                    self.error("\n")
                    traceback.print_tb(e.__traceback__, file=self.stderr)
                    self.error("\n")
        finally:
            self.exit_grammar()

    def process_cli_response(self, tokens, cli_response):
        if tokens and cli_response.result != MATCH_SUCCESS:
            self.error("Result:", cli_response.result)
            self.error("Error:", cli_response.error)



# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import sys
from io import StringIO
from contextlib import contextmanager


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err




# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import sys


ESCAPED_CHAR_INPUTS = [
    '\\n',
    '\\0',
    '\\r',
    '\\t',
    '\\b',
    '\\v',
    '\\a',
    '\\"',
    '\\\\'
]


class ExtendedString(str):

    def __new__(cls, value, *args, **kwargs):
    	return super(ExtendedString, cls).__new__(cls, value)

    def __init__(self, value, *args, **kwargs):
        self._args = args
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def value(self):
        return str(self)


def convert_to_python_string(cli_string):

    converted_str = cli_string

    if hasattr(cli_string, "quoted") and cli_string.quoted:
        if converted_str.startswith('"'):
            converted_str = converted_str[1:]

        if not hasattr(cli_string, "quote_incomplete") or not cli_string.quote_incomplete:
            if converted_str.endswith('"'):
                converted_str = converted_str[:-1]

    if any(pattern in converted_str for pattern in ESCAPED_CHAR_INPUTS):
        parts = converted_str.split("\\\\")
        converted_parts = []
        for part in parts:
            part = part.replace("\\n", "\n")
            part = part.replace("\\0", "\0")
            part = part.replace("\\r", "\r")
            part = part.replace("\\t", "\t")
            part = part.replace("\\b", "\b")
            part = part.replace("\\v", "\v")
            part = part.replace("\\a", "\a")
            part = part.replace('\\"', '"')
            converted_parts.append(part)

        converted_str = "\\".join(converted_parts)
    return converted_str


def convert_to_cli_string(python_string):
    if any(c in python_string for c in ["\n", "\0", "\r", "\t", "\b", "\a", "\v", '"']):
        parts = python_string.split("\\")
        converted_parts = []
        for part in parts:
            part = part.replace("\n", "\\n")
            part = part.replace("\0", "\\0")
            part = part.replace("\r", "\\r")
            part = part.replace("\t", "\\t")
            part = part.replace("\b", "\\b")
            part = part.replace("\v", "\\v")
            part = part.replace("\a", "\\a")
            part = part.replace('"', '\\"')
            converted_parts.append(part)

        return ExtendedString('"' + "\\\\".join(converted_parts) + '"', quoted=True)
    return python_string


class StdStreamsHolder():

    def init_streams(self, stdin=None, stdout=None, stderr=None):
        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr

    @property
    def stdin(self):
        return self._stdin if self._stdin is not None else sys.stdin

    @property
    def stdout(self):
        return self._stdout if self._stdout is not None else sys.stdout

    @property
    def stderr(self):
        return self._stderr if self._stderr is not None else sys.stderr

    def print(self, *args):
        print(*args, file=self.stdout)

    def error(self, *args):
        print(*args, file=self.stderr)




# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import unittest


from nessaid_cli_tests.test_tokenizer import tokenizer_test
from nessaid_cli_tests.test_cli import cli_test
from nessaid_cli_tests.test_grammar import grammar_test


def doTests():
  print('Started Nessaid Cli Python implementation testing.\n')
  unittest.TextTestRunner(verbosity=2).run(tokenizer_test)
  unittest.TextTestRunner(verbosity=2).run(cli_test)
  unittest.TextTestRunner(verbosity=2).run(grammar_test)


# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import sys
import inspect
import asyncio
import unittest

from nessaid_cli.cmd import NessaidCmd
from nessaid_cli.tokenizer.tokenizer import NessaidCliTokenizer

from nessaid_cli.tokens import (
    StringToken,
    RangedIntToken,
    BooleanToken,
    RangedStringToken,
)

from nessaid_cli_tests.test_utils import captured_output


class Cmd1(NessaidCmd):
    """
    token STRING_TOKEN StringToken();
    token BOOLEAN BooleanToken();
    token RANGED_STRING_TOKEN_1 RangedStringToken(5, 10);
    token RANGED_INT_TOKEN_1 RangedIntToken(0, 100);
    token RANGED_INT_TOKEN_2 RangedIntToken(1, 1000);
    token RANGED_INT_TOKEN_3 RangedIntToken(-100, -1);
    token RANGED_INT_TOKEN_4 RangedIntToken(-100, 100);
    token RANGED_INT_TOKEN_5 RangedIntToken(-100, 0);
    """

    def get_token_classes(self):
        return [
            StringToken,
            RangedStringToken,
            RangedIntToken,
            BooleanToken,
        ]

    def do_basic_1(self, cli_input, cli_output):
        """
        "input" << $cli_input = $1; $cli_output = "output"; >>
        """
        print("input:", cli_input)
        print("output:", cli_output)

    def do_types_1(self, cli_input, cli_output):
        """
        "type"
        (
            (
                "string"
                STRING_TOKEN
                << $cli_input = $1; $cli_output = $2; >>
            )
            |
            (
                "ranged-string"
                RANGED_STRING_TOKEN_1
                << $cli_input = $1; $cli_output = $2; >>
            )
            |
            (
                "boolean"
                BOOLEAN
                << $cli_input = $1; $cli_output = $2; >>
            )
            |
            (
                "int"
                RANGED_INT_TOKEN_1
                << $cli_input = $1; $cli_output = $2; >>
            )
            |
            (
                "bigger-int"
                RANGED_INT_TOKEN_2
                << $cli_input = $1; $cli_output = $2; >>
            )
            |
            (
                "negative-int"
                RANGED_INT_TOKEN_3
                << $cli_input = $1; $cli_output = $2; >>
            )
            |
            (
                "negative-or-positive"
                RANGED_INT_TOKEN_4
                << $cli_input = $1; $cli_output = $2; >>
            )
            |
            (
                "negative-or-zero"
                RANGED_INT_TOKEN_5
                << $cli_input = $1; $cli_output = $2; >>
            )
        )
        """
        print("Input:", cli_input)
        print("Type:", type(cli_output))
        print("Output:", cli_output)

class CmdTest1(unittest.TestCase):

    def test_basic_1(self):
        input_output_err = [
            (
                [""],
                [""],
                [""]
            ),
            (
                ["   "],
                [""],
                [""]
            ),
            (
                ["   ", "   ", "   "],
                [""],
                [""]
            ),
            (
                ["input"],
                ["input: input", "output: output"],
                [""]
            ),
        ]
        loop = asyncio.get_event_loop()
        for inp, out, err in input_output_err:
            with captured_output() as (stdout, stderr):
                loop.run_until_complete(Cmd1.execute_args(*inp))
            stdout = stdout.getvalue().strip().split("\n")
            stderr = stderr.getvalue().strip().split("\n")

            assert out == stdout, "\nstdout: Expected: {}\nstdout: Actual  : {}".format(out, stdout)
            assert err == stderr, "\nstderr: Expected: {}\nstderr: Actual  : {}".format(err, stderr)

    def do_test_type_positive(self, python_type, cli_prefixes, cli_type, cli_types, type_inputs):
        loop = asyncio.get_event_loop()
        cmd = Cmd1(prompt="# ")

        cases = []
        count = 0

        for p in cli_prefixes:
            for c in cli_types:
                for i, o in type_inputs:
                    count += 1
                    inp = " ".join([p, c, i])
                    out = "Input: {}\nType: <class '{}'>\nOutput: {}".format(cli_type, python_type, o).strip()
                    #if count == 100:
                    #    out += "123"
                    err = ""
                    with captured_output() as (stdout, stderr):
                        loop.run_until_complete(cmd.execute_line(inp))

                    stdout = stdout.getvalue().strip()
                    stderr = stderr.getvalue().strip()

                    info = "\n\nFunction: {}".format(inspect.stack()[0][3])
                    info += "\ncli_prefix: {}".format(p)
                    info += "\npython_type: {}".format(python_type)
                    info += "\ncli_type: {}".format(cli_type)
                    info += "\ncli_type input: {}".format(c)
                    info += "\ntype value input: {}".format(i)
                    info += "\ntype value output input: {}".format(o)

                    assert out == stdout, info + "\n\nstdout: Expected: {}\nstdout: Actual  : {}".format(out, stdout)
                    assert err == stderr, info + "\nstderr: Expected: {}\nstderr: Actual  : {}".format(err, stderr)

    def do_test_type_partial(self, python_type, cli_prefixes, cli_type, cli_types, type_inputs):
        loop = asyncio.get_event_loop()
        cmd = Cmd1(prompt="# ")
        cases = []
        count = 0

        for p in cli_prefixes:
            for c in cli_types:
                for i in type_inputs:
                    count += 1
                    inp = " ".join([p, c, i])
                    out = ""
                    err = "Result: partial\nError: Input sequence is not complete"
                    with captured_output() as (stdout, stderr):
                        loop.run_until_complete(cmd.execute_line(inp))

                    stdout = stdout.getvalue().strip()
                    stderr = stderr.getvalue().strip()

                    info = "\n\nFunction: {}".format(inspect.stack()[0][3])
                    info += "\ncli_prefix: {}".format(p)
                    info += "\npython_type: {}".format(python_type)
                    info += "\ncli_type: {}".format(cli_type)
                    info += "\ncli_type input: {}".format(c)
                    info += "\ntype value input: {}".format(i)

                    stdout_info = "\n\nstdout: Expected:\n{}\nstdout: Actual  :\n{}\n".format(out, stdout)
                    stderr_info = "\nstderr: Expected:\n{}\nstderr: Actual  :\n{}\n".format(err, stderr)

                    assert out == stdout, "Failure: stdout mismatch:\n\n" + info + stdout_info + stderr_info
                    assert err == stderr, "Failure: stderr mismatch:\n\n" + info + stdout_info + stderr_info

    def do_test_type_negative(self, python_type, cli_prefixes, cli_type, cli_types, type_inputs):
        loop = asyncio.get_event_loop()
        cmd = Cmd1(prompt="# ")
        cases = []
        count = 0

        for p in cli_prefixes:
            for c in cli_types:
                for i in type_inputs:
                    count += 1
                    inp = " ".join([p, c, i])
                    out = ""
                    err = "Result: failure\nError: Could not match any rule for this sequence"
                    with captured_output() as (stdout, stderr):
                        loop.run_until_complete(cmd.execute_line(inp))

                    stdout = stdout.getvalue().strip()
                    stderr = stderr.getvalue().strip()

                    info = "\n\nFunction: {}".format(inspect.stack()[0][3])
                    info += "\ncli_prefix: {}".format(p)
                    info += "\npython_type: {}".format(python_type)
                    info += "\ncli_type: {}".format(cli_type)
                    info += "\ncli_type input: {}".format(c)
                    info += "\ntype value input: {}".format(i)

                    stdout_info = "\n\nstdout: Expected:\n{}\nstdout: Actual  :\n{}\n".format(out, stdout)
                    stderr_info = "\nstderr: Expected:\n{}\nstderr: Actual  :\n{}\n".format(err, stderr)

                    assert out == stdout, "Failure: stdout mismatch:\n\n" + info + stdout_info + stderr_info
                    assert err == stderr, "Failure: stderr mismatch:\n\n" + info + stdout_info + stderr_info

    def test_type_string_positive(self):
        self.do_test_type_positive(
            "str",
            ["t", "ty", "typ", "type"],
            "string",
            ["s", "st", "str", "string"],
            [
                ('asdfgh', 'asdfgh'),
                ('"asdfgh"', 'asdfgh'),
                ('"type string"', 'type string'),
                (r'"type\nstring"', 'type\nstring'),
                (r'"type\\nstring"', 'type\\nstring'),
                (r'"type\tstring"', 'type\tstring'),
                (r'"type\\tstring"', 'type\\tstring'),
                (r'"type\\string"', 'type\\string'),
                (r'"type\\\\string"', 'type\\\\string'),
                ('""', ''),
                ('12345', '12345'),
                ('"12345"', '12345'),
            ]
        )

    def test_type_string_partial(self):
        self.do_test_type_partial(
            "str",
            ["t", "ty", "typ", "type"],
            "string",
            ["s", "st", "str", "string"],
            ["", " "]
        )

    def test_type_string_negative(self):
        self.do_test_type_negative(
            "str",
            ["t", "ty", "typ", "type"],
            "string",
            ["s", "st", "str", "string"],
            ["a b"]
        )


testcase1 = unittest.TestLoader().loadTestsFromTestCase(CmdTest1)

cli_test = unittest.TestSuite([testcase1])


# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import os
import string
import fnmatch
import platform
from pathlib import Path

from nessaid_cli.utils import (
    ESCAPED_CHAR_INPUTS,
    convert_to_cli_string,
    convert_to_python_string
)


MATCH_SUCCESS = 'success'
MATCH_FAILURE = 'failure'
MATCH_PARTIAL = 'partial'
MATCH_AMBIGUOUS = 'ambigous'

TOO_MANY_COMPLETIONS = -1


class CliToken():

    def __init__(self, name):
        self._name = name
        self._cli_string = convert_to_cli_string(name)

    def __repr__(self):
        return self._cli_string

    def __str__(self):
        return self.__repr__()

    @property
    def case_insensitive(self):
        return False

    def get_value(self, match_string=None):
        if self.completable:
            _, completions = self.complete(match_string)
            if len(completions) == 1:
                return convert_to_python_string(completions[0])
            elif self.match(match_string) == MATCH_SUCCESS:
                return convert_to_python_string(match_string)
        else:
            if self.match(match_string) == MATCH_SUCCESS:
                return convert_to_python_string(match_string)
        raise ValueError("Fix matching logic for this token: {}".format(self.__class__.__name__))

    @property
    def name(self):
        return self._name

    @property
    def helpstring(self):
        return self._cli_string

    @property
    def completable(self):
        return True

    def match(self, token_input):
        if self._cli_string == token_input:
            return MATCH_SUCCESS
        elif self._cli_string.startswith(token_input):
            return MATCH_PARTIAL
        else:
            return MATCH_FAILURE

    def complete(self, token_input):
        if not token_input:
            return 1, [self._cli_string]
        elif self._cli_string.startswith(token_input):
            return 1, [self._cli_string]
        else:
            return 0, []


class AlternativeStringsToken(CliToken):

    def __init__(self, name, alternatives, *args):
        super().__init__(name)
        if (isinstance(alternatives, list) or
            isinstance(alternatives, set) or
            isinstance(alternatives, tuple)):
            self._alternatives = list(alternatives)
        elif args:
            self._alternatives = [alternatives] + list(args)
        else:
            self._alternatives = [alternatives]
        self._cli_strings = [convert_to_cli_string(s) for s in self._alternatives]

    @property
    def helpstring(self):
        return "Any one of: {}".format(set(self._cli_strings))

    @property
    def completable(self):
        return True

    def complete(self, token_input):
        if not token_input:
            return len(self._cli_strings), list(self._cli_strings)
        completions = set()
        for e in self._cli_strings:
            if token_input and e.startswith(token_input):
                completions.add(e)
        return len(completions), list(completions)

    def get_value(self, match_string=None):
        v = super().get_value(match_string)
        return v

    def match(self, token_input):
        if token_input and token_input in self._cli_strings:
            return MATCH_SUCCESS
        n, completions = self.complete(token_input)
        if n == TOO_MANY_COMPLETIONS:
            return MATCH_PARTIAL
        if not completions:
            return MATCH_FAILURE
        elif len(completions) == 1:
            return MATCH_SUCCESS
        else:
            return MATCH_PARTIAL


class StringToken(CliToken):

    def __init__(self, name):
        super().__init__(name)

    @property
    def helpstring(self):
        return "Any string"

    @property
    def completable(self):
        return False

    def complete(self, token_input):
        return TOO_MANY_COMPLETIONS, []

    def match(self, token_input):
        return MATCH_PARTIAL

    def get_value(self, match_string=None):
        if isinstance(match_string, str):
            if match_string.startswith('"'):
                match_string = match_string[1:]
            if match_string.endswith('"'):
                match_string = match_string[:-1]
            return match_string
        return None


class RangedStringToken(StringToken):

    def __init__(self, name, min_len, max_len):
        min
        super().__init__(name)
        self._min_len = min(int(min_len), int(max_len))
        self._max_len = max(int(min_len), int(max_len))

        if (self._min_len < 0) or (self._max_len < 0):
            raise ValueError("Negative size")

    @property
    def helpstring(self):
        return "Any string of length ({}-{})".format(self._min_len, self._max_len)

    @property
    def completable(self):
        return False

    def complete(self, token_input):
        return TOO_MANY_COMPLETIONS, []

    def get_value(self, match_string=None):
        s = super().get_value(match_string)
        if isinstance(s, str):
            if len(s) >= self._min_len and len(s) <= self._max_len:
                return s
        return None

    def match(self, token_input):
        val = super().get_value(token_input)
        if len(val) > self._max_len:
            return MATCH_FAILURE
        if len(val) == self._max_len:
            return MATCH_SUCCESS
        return MATCH_PARTIAL


class RangedIntToken(CliToken):

    def __init__(self, name, start, end, max_suggestions=10):
        start = int(start)
        end = int(end)
        super().__init__(name)
        self._start = min(start, end)
        self._end = max(start, end)
        self._max_suggestions = max_suggestions

    @property
    def helpstring(self):
        return "An integer between {} and {}".format(self._start, self._end)

    def get_value(self, match_string=None):
        try:
            number = int(match_string)
            if number >= self._start and number <= self._end:
                return number
        except Exception:
            pass
        return None

    @property
    def completable(self):
        return True

    def _complete(self, min_limit, max_limit, number):

        if number == 0:
            if (max_limit - min_limit + 1) > self._max_suggestions:
                return max_limit - min_limit + 1, []
            return max_limit - min_limit + 1, list(range(min_limit, max_limit + 1))

        if number > max_limit:
            return 0, []

        min_len = len(str(min_limit))
        max_len = len(str(max_limit))
        number_len = len(str(number))

        count = 0
        completions = []
        min_num = number
        max_num = number
        num_len = number_len
        power = 10

        if num_len < min_len:
            while num_len < min_len:
                min_num = min_num * 10
                max_num = min_num + power - 1
                num_len += 1
                power *= 10

        if num_len == min_len:
            if num_len == max_len and (max_num < min_limit or min_num > max_limit):
                return 0, []
            else:
                lower_limit = max(min_limit, min_num)
                upper_limit = min(max_limit, max_num)
                count = len(range(lower_limit, upper_limit + 1))
                if count > self._max_suggestions:
                    return count, []
                completions = list(range(lower_limit, upper_limit + 1))
                if num_len == max_len or count > self._max_suggestions:
                    return count, completions
        else:
            count = 1
            completions = [number]

        num_len = num_len + 1
        min_num = min_num * 10
        max_num = min_num + power - 1

        while num_len < max_len:
            count += power
            if count > self._max_suggestions:
                return count, []

            completions += list(range(min_num, max_num + 1))
            num_len += 1
            power = 10 ** (num_len - number_len)
            min_num = min_num * 10
            max_num = min_num + power - 1

        if min_num <= max_limit and num_len == max_len:
            count += len(range(min_num, min(max_num, max_limit) + 1))
            if count > self._max_suggestions:
                return count, []
            completions += list(range(min_num, min(max_num, max_limit) + 1))

        return count, completions

    def match(self, token_input):
        if isinstance(token_input, str):
            n, comps = self.complete(token_input)
            if n > 1 or n == TOO_MANY_COMPLETIONS:
                return MATCH_PARTIAL
            elif n == 1:
                return MATCH_SUCCESS
        return MATCH_FAILURE

    def complete(self, token_input):
        if isinstance(token_input, str) and str:
            count = 0
            comps = []

            if token_input == '-':
                if self._start >= 0:
                    return 0, []
                if self._end >= 0:
                    comp_args = [0, -self._start, 0]
                else:
                    comp_args = [-self._end, -self._start, 0]
                n, comps = self._complete(*comp_args)
                if n > 0 and not comps:
                    return TOO_MANY_COMPLETIONS, []
                return n, [-c for c in comps]
            elif token_input == '':
                count = self._end - self._start + 1
                if count > 10:
                    return count, []
                elif count == 1:
                    return 1, [str(self._start)]
                elif count == 0:
                    return 0, []
                if count > self._max_suggestions:
                    comps = []
                else:
                    comps = list(range(self._start, self._end + 1))
                return count, comps

            else:
                try:
                    number = int(token_input)
                except Exception:
                    return 0, []

                if number < 0 or token_input.startswith("-"):
                    if number == 0 and self._start == 0:
                        return 1, ["0"]
                    if self._start >= 0:
                        return 0, []
                    if self._end >= 0:
                        comp_args = [0, -self._start, -number]
                    else:
                        comp_args = [-self._end, -self._start, -number]

                    n, comps = self._complete(*comp_args)
                    if n > 0 and not comps:
                        return TOO_MANY_COMPLETIONS, []
                    return n, [str(-c) for c in comps]
                else:
                    if self._end < 0:
                        return 0, []

                    if self._start <= 0:
                        comp_args = [0, self._end, number]
                    else:
                        comp_args = [self._start, self._end, number]
                    n, comps = self._complete(*comp_args)
                    if n > 0 and not comps:
                        return TOO_MANY_COMPLETIONS, []
                    return n,  [str(c) for c in comps]


class RangedDecimalToken(CliToken):

    def __init__(self, name, start, end):
        start = float(start)
        end = float(end)
        super().__init__(name)
        self._start = min(start, end)
        self._end = max(start, end)

    @property
    def helpstring(self):
        return "A decimal number between {} and {}".format(self._start, self._end)

    @property
    def completable(self):
        return False

    def complete(self, token_input):
        return 0, []

    def match(self, token_input):
        if isinstance(token_input, str):
            if token_input == "":
                return MATCH_PARTIAL
            if token_input == "-":
                if self._start >= 0 :
                    return MATCH_FAILURE
                return MATCH_PARTIAL
            try:
                decimal = float(token_input)
            except Exception:
                return MATCH_FAILURE
            if decimal > 0 and decimal > self._end:
                return MATCH_FAILURE
            if decimal < 0 and decimal < self._start:
                return MATCH_FAILURE
            return MATCH_PARTIAL
        return MATCH_FAILURE

    def get_value(self, match_string=None):
        try:
            number = float(match_string)
            if number >= self._start and number <= self._end:
                return number
        except Exception:
            pass
        return None



class PathTokenPath():

    def __init__(self, path, path_string, partial=False):
        self.path = path
        self.path_string = path_string
        self.partial = partial
        self.is_dir = path.is_dir()
        self.exists = path.exists()
        self.has_dir_completion = False

    def __repr__(self):
        return "{}: {}".format(self.path, self.path_string)

    def __str__(self):
        return "{}: {}".format(self.path, self.path_string)


class BooleanToken(StringToken):

    def __init__(self, name):
        super().__init__(name)

    @property
    def completable(self):
        return True

    @property
    def case_insensitive(self):
        return True

    @property
    def helpstring(self):
        return "True or False"

    def complete(self, str_input):
        if not str_input:
            return 2, ["True", "False"]
        elif str_input.lower() in ["true", "false"]:
            return 1, [str_input]
        elif "true".startswith(str_input.lower()):
            return 1, [str_input + "true"[len(str_input):]]
        elif "false".startswith(str_input.lower()):
            return 1, [str_input + "false"[len(str_input):]]
        return 0, []

    def match(self, str_input):
        n, l = self.complete(str_input)
        if n == 0:
            return MATCH_FAILURE
        elif n == 1:
            if str_input.lower() in ["true", "false"]:
                return MATCH_SUCCESS
        return MATCH_PARTIAL

    def get_value(self, str_input):
        n, l = self.complete(str_input)
        if n == 1:
            if l[0].lower() == "true":
                return True
            return False
        return None


class PathToken(StringToken):

    ANY = 'path'
    FILE = 'file'
    DIRECTORY = 'directory'

    def __init__(self, name, pathtype=ANY):
        self._pathtype = pathtype
        self._is_windows = None
        super().__init__(name)

    def get_drives(self):
        drives = []
        if self.is_windows:
            for c in string.ascii_lowercase:
                if os.path.isdir(c + ':'):
                    drives.append(c + ':')
        return drives

    @property
    def completable(self):
        return True

    @property
    def is_windows(self):
        if self._is_windows is None:
            os_system = platform.system()
            if os_system.lower() == 'windows':
                self._is_windows = True
            else:
                self._is_windows = False
        return self._is_windows

    @property
    def case_insensitive(self):
        if self.is_windows:
            return True
        return False

    @property
    def helpstring(self):
        if self.is_windows:
            path_sep = "\\\\"
        else:
            path_sep = "/"
        return "A {}.".format(self._pathtype) + ' Start the input with quote (") and use {} as separator'.format(path_sep)

    def complete(self, str_input):
        return TOO_MANY_COMPLETIONS, []

    def children(self, path):
        try:
            content = os.listdir(path)
            if "." in content:
                content.remove(".")
            if ".." in content:
                content.remove("..")
            return content
        except PermissionError:
            print("\nPermissionError on {}\n".format(path))
            return []

    def get_value(self, str_input):
        m, n, l, _ = self.lookup(str_input)
        if n == 0:
            return None
        elif n == 1:
            return l[0].path_string
        else:
            return str_input

    def complete(self, str_input):
        if str_input == "":
            return TOO_MANY_COMPLETIONS, []
        elif str_input == '"':
            if self.is_windows:
                path_sep = "\\\\"
            else:
                path_sep = "/"

            options = ['".', '"..', '"' + path_sep]
            children = self.children(os.path.curdir)
            options += ['"' + c for c in children]
            return len(options), options
        else:
            if str_input.startswith('"') and str_input.endswith('"') and len(str_input) > 1:
                path_complete = True
            else:
                path_complete = False
            m, n, l, path_sep = self.lookup(str_input)

            if not path_complete:
                for elem in l.copy():
                    if elem.is_dir and elem.has_dir_completion:
                        children = self.children(elem.path)
                        for c in children:
                            path_string = elem.path_string + c
                            l += [PathTokenPath(Path(path_string), path_string)]

            if m == MATCH_PARTIAL and len(l) == 1:
                if l[0].is_dir and not l[0].has_dir_completion:
                    l.append(PathTokenPath(l[0].path, l[0].path_string + path_sep, partial=True))
                if not l[0].is_dir:
                    path_complete = True
                else:
                    if not l[0].has_dir_completion:
                        if l[0].path_string == os.path.curdir or l[0].path_string.endswith(path_sep + os.path.curdir):
                            p = l[0].path_string + os.path.curdir
                            l.append(PathTokenPath(Path(p), p, partial=True))
                    else:
                        if not self.children(l[0].path):
                            path_complete = True

            return n, ['"' + str(elem.path_string) + ('"' if path_complete else "") for elem in l]

    def match(self, str_input):

        if str_input.startswith('"') and str_input.endswith('"') and len(str_input) > 1:
            path_complete = True
        else:
            path_complete = False
        m, n, l, _ = self.lookup(str_input)
        if m == MATCH_PARTIAL and len(l) == 1:
            if path_complete:
                return MATCH_SUCCESS
            elif not l[0].is_dir:
                return MATCH_SUCCESS
        return m

    def lookup(self, str_input):
        path_complete = False
        if str_input == "":
            return MATCH_PARTIAL, TOO_MANY_COMPLETIONS, [], os.path.pathsep

        if str_input.startswith('"'):
            str_input = str_input[1:]

            if str_input.endswith('"'):
                str_input = str_input[:-1]
                path_complete = True

        segments = str_input.split(os.path.sep)
        add_drive_mark = ""

        path_sep = "/"
        if not segments:
            if not self.is_windows:
                return MATCH_FAILURE, 0, [], path_sep
            segments = str_input.split("/")
            if not segments:
                return MATCH_FAILURE, 0, [], path_sep
        else:
            if self.is_windows:
                path_sep = "\\"

        path_string = ""
        partial_drive = None

        if segments[0] == "":
            if len(segments) == 1:
                path_string = ""
            else:
                path_string = path_sep
        elif self.is_windows:
            if len(segments[0]) == 2 and segments[0][1] == ":":
                path_string = segments[0]
                add_drive_mark = path_sep
            elif len(segments[0]) == 3 and segments[0][0] == "/" and segments[0][2] == ":":
                path_string = segments[0][1:]
                add_drive_mark = path_sep

        if path_string:
            path_objects = [PathTokenPath(Path(path_string + add_drive_mark), path_string)]
            if path_string == path_sep:
                path_objects[0].has_dir_completion = True
            segments.pop(0)
        else:
            path_objects = [PathTokenPath(Path(os.path.curdir), "")]
            if self.is_windows and len(segments) == 1 and not path_complete:
                if (segments[0].lower() + ":") in self.get_drives():
                    partial_drive = Path(segments[0] + ":\\")
                    partial_drive_str = segments[0] + ":"

        while segments:
            path_objects = [p for p in path_objects if p.is_dir and not p.partial]
            for p in path_objects:
                if p.path_string and not p.path_string.endswith(path_sep):
                    p.path_string += path_sep
                    p.has_dir_completion = True

            segment = segments.pop(0)
            if segment == "":
                break

            p.has_dir_completion = False

            if segment == ".":
                for p in path_objects:
                    p.path_string += "."
                continue
            elif segment == "..":
                for p in path_objects:
                    p.path_string += ".."
                    p.path = Path(p.path_string)
            elif segment == "*":
                opts = []
                for p in path_objects:
                    children = self.children(p.path)
                    for c in children:
                        path_string = p.path_string + c
                        opts.append(PathTokenPath(Path(path_string), path_string))
                path_objects = opts
            else:
                opts = []
                for p in path_objects:
                    children = self.children(p.path)
                    if self.case_insensitive:
                        exacts = [p.path_string + c for c in children if c.lower() == segment.lower()]
                    else:
                        exacts = [p.path_string + c for c in children if c == segment]

                    matches = [p.path_string + c for c in children if fnmatch.fnmatch(c, segment) and p.path_string + c not in exacts]
                    matches = list(set(matches + exacts))
                    opts += [PathTokenPath(Path(m), m) for m in matches]
                    if not path_complete:
                        if not self.case_insensitive:
                            path_strings = matches
                            partial_matches = [PathTokenPath(Path(p.path_string + c), p.path_string + c, partial=True) for c in children if c.startswith(segment) and p.path_string + c not in path_strings]
                        else:
                            path_strings = [m.lower() for m in matches]
                            partial_matches = [PathTokenPath(Path(p.path_string + c), p.path_string + c, partial=True) for c in children if c.lower().startswith(segment.lower()) and p.path_string.lower() + c.lower() not in path_strings]
                        opts += partial_matches

                if partial_drive:
                    path_objects = [PathTokenPath(partial_drive, partial_drive_str, partial=True)]
                    partial_drive = None
                else:
                    path_objects = []
                path_objects += opts

        dirs = [d for d in path_objects if d.is_dir]
        files = [f for f in path_objects if not f.is_dir]

        if (len(dirs) + len(files)) == 1:
            if (dirs + files)[0].partial or not path_complete:
                return MATCH_PARTIAL, 1, dirs + files, path_sep
            else:
                return MATCH_SUCCESS, 1, dirs + files, path_sep

        if (len(dirs) + len(files)) == 0:
            return MATCH_FAILURE, 0, [], path_sep
        return MATCH_PARTIAL, len(dirs) + len(files), (dirs + files), path_sep



# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#


from nessaid_cli.lex_yacc_common import (
    NessaidCliLexerCommon,
    NessaidCliParserCommon,
)

from nessaid_cli.elements import (
    ConstantInputElement,
    KeywordInputElement,
    SequenceInputElement,
    OptionalInputElement,
    AlternativeInputElement,
    NamedGrammar,
    GrammarRefElement,
    UnresolvedInputElement,
    TokenClassDef,
    GrammarSpecification,
    CliParameter,
    CliArgument,
    CliTokenArgument
)

from nessaid_cli.binding_parser.binding_text_parser import NessaidCliBindingParser

from nessaid_cli.utils import convert_to_python_string


class NessaidCliLexer(NessaidCliLexerCommon):

    BINDING_STATE = 'BINDING'

    states = (
        (BINDING_STATE, 'exclusive'),
    )

    tokens = NessaidCliLexerCommon.tokens + (
        'IMPORT',
        'TOKEN',
        'OPEN_BINDING',
        'CLOSE_BINDING',
        'BOUND_CONTENT',
        'TRUE',
        'FALSE',
        'NONE',
    )

    t_BINDING_ignore = ""

    def t_IDENTIFIER(self, t):
        r'[A-Za-z_][-a-zA-Z0-9_]*'
        if t.value in NessaidCliLexer.KYWORD_TOKENS:
            t.type = NessaidCliLexer.KYWORD_TOKENS[t.value]
        return t

    """
    def t_IMPORT(self, t):
        "import"
        return t

    def t_TOKEN(self, t):
        "token"
        return t

    def t_TRUE(self, t):
        "True"
        return t

    def t_FALSE(self, t):
        "False"
        return t

    def t_NONE(self, t):
        "None"
        return t
    """

    KYWORD_TOKENS = {
        "import": 'IMPORT',
        "token": 'TOKEN',
        "True": 'TRUE',
        "False": 'FALSE',
        "None": 'NONE',
    }

    t_ignore_COMMENT = NessaidCliLexerCommon.common_COMMENT

    def t_INITIAL_OPEN_BINDING(self, t):
        r'<<'
        self.enter_state(NessaidCliLexer.BINDING_STATE)
        return t

    def t_BINDING_BOUND_CONTENT(self, t):
        r'([^>\\])+'
        self.update_counters(t)
        return t

    def t_BINDING_ESCAPED_CHAR(self, t):
        r'\\(.)'
        self.update_counters(t)
        return t

    def t_BINDING_CLOSE_BINDING(self, t):
        r'>>'
        self.exit_state(NessaidCliLexer.BINDING_STATE)
        return t

    t_BINDING_error = NessaidCliLexerCommon.t_error

    def __init__(self, stdin=None, stdout=None, stderr=None):
        # Build the lexer
        super().__init__(stdin=stdin, stdout=stdout, stderr=stderr)


class NessaidCliParser(NessaidCliParserCommon):

    tokens = NessaidCliLexer.tokens

    def p_grammar_spec(self, t):
        """grammar_spec : content empty"""
        content = t[1]
        grammar_spec = self._grammar_spec
        t[0] = grammar_spec

    def p_content(self, t):
        """content : content block
                   | empty"""
        content = t[1]
        if content is None:
            content = []
        else:
            block = t[2]
            content.append(block)
        t[0] = content

    def p_block(self, t):
        """block : named_grammar
                 | token_block
                 | unused_token"""
        grammar = t[1]
        block = grammar
        t[0] = block

    def p_token_block(self, t):
        '''token_block : TOKEN token_spec_list SEMICOLON'''
        token_spec_list = t[2]
        for token_spec in token_spec_list:
            token_name = token_spec[0]
            token_classdef = token_spec[1]
            self._grammar_spec.add_token_def(token_name, token_classdef)

    def p_token_spec_list(self, t):
        """token_spec_list : token_spec_list COMMA token_unit
                           | token_unit"""
        if len(t) == 2:
            token_spec_list = [t[1]]
        else:
            token_spec_list = t[1]
            token_spec = t[3]
            token_spec_list.append(token_spec)
        t[0] = token_spec_list

    def p_token_unit(self, t):
        '''token_unit : identifier optional_class_def'''

        token_name = t[1]
        classdef = t[2]
        token_spec = (token_name, classdef)
        t[0] = token_spec

    def p_optional_class_def(self, t):
        """optional_class_def : identifier LPAREN optional_token_arguments RPAREN
                              | empty"""
        classname = t[1]
        if classname is None:
            classname =  "CliToken"
            arglist = []
        else:
            arglist = t[3]
            if arglist is None:
                arglist = []
        classdef = TokenClassDef(classname, arglist)
        t[0] = classdef

    def p_named_grammar(self, t):
        'named_grammar : identifier optional_parameter_list COLON rule SEMICOLON'
        identifier = t[1]
        param_list = t[2]
        rule = t[4]
        named_grammar = NamedGrammar(identifier, param_list, rule)
        rule.parent = named_grammar
        self._grammar_spec.add_grammar(named_grammar)
        t[0] = named_grammar

    def p_term_multiplier(self, t):
        """term : term MULTIPLY repeater"""
        term = t[1]
        repeater = t[3]
        min_count = repeater[0]
        max_count = repeater[1]

        if min_count < 0 or max_count < 0:
            raise ValueError("Only positive repeaters allowed")

        if min_count == max_count:
            if min_count == 0:
                raise ValueError("0 repeaters not allowed. Delete the term")
            if min_count == 1:
                t[0] = term
            else:
                t[0] = SequenceInputElement((term,))
                term.parent = t[0]
                term.position = 0
                t[0].repeat_count = min_count
            return
        elif min_count == 0:
            if max_count == 1:
                t[0] = OptionalInputElement((term,))
                term.parent = t[0]
            else:
                t0 = SequenceInputElement((term,))
                term.parent = t0
                term.position = 0
                t0.repeat_count = max_count
                t[0] = OptionalInputElement((t0,))
                t0.parent = t[0]
            return

        term_copy = term.copy()
        if min_count == 1:
            block_1 = term
        else:
            block_1 = SequenceInputElement((term,))
            term.parent = block_1
            term.position = 0
            block_1.repeat_count = min_count

        opt_block = OptionalInputElement((term_copy,))
        term_copy.parent = opt_block

        block_2 = SequenceInputElement((opt_block,))
        block_2.repeat_count = max_count - min_count
        opt_block.parent = block_2
        opt_block.position = 0

        t[0] = SequenceInputElement((block_1, block_2))
        block_1.parent = t[0]
        block_2.parent = t[0]

    def p_repeater(self, t):
        """repeater : INTEGER
                    | LPAREN INTEGER RPAREN
                    | LPAREN INTEGER COLON INTEGER RPAREN"""
        if len(t) == 2:
            repeater = t[1]
            repeater_range = (int(repeater), int(repeater))
        elif len(t) == 4:
            repeater = t[2]
            repeater_range = (int(repeater), int(repeater))
        else:
            rmin = int(t[2])
            rmax = int(t[4])
            repeater_range = (min(rmin, rmax), max(rmin, rmax))
        t[0] = repeater_range

    def p_parenthesised_rule(self, t):
        """term : LPAREN rule RPAREN"""
        rule = t[2]
        t[0] = rule

    def p_optional_rule(self, t):
        """term : LBRACE rule RBRACE"""
        rule = t[2]
        if type(rule) is SequenceInputElement:
            values = rule.value
            pre_match_binding = rule.pre_match_binding
            post_match_binding = rule.post_match_binding
            opt_rule = OptionalInputElement(values)
            opt_rule.pre_match_binding = pre_match_binding
            opt_rule.post_match_binding = post_match_binding
        else:
            values = (rule, )
            opt_rule = OptionalInputElement(values)

        for elem in values:
            elem.parent = opt_rule

        t[0] = opt_rule

    def p_rule(self, t):
        """rule : unit_with_binding
                | rule_alternatives"""
        rule = t[1]
        t[0] = rule

    def p_rule_or_rule(self, t):
        """rule_alternatives : unit_with_binding OR unit_with_binding"""
        unit_1 = t[1]
        unit_2 = t[3]
        values = (unit_1, unit_2)
        alt = AlternativeInputElement(values)
        for elem in values:
            elem.parent = alt
        t[0] = alt

    def p_alternatives_or_rule(self, t):
        """rule_alternatives : rule_alternatives OR unit_with_binding"""
        alt = t[1]
        unit = t[3]
        values = alt.value + (unit, )
        alt = AlternativeInputElement(values)
        for elem in values:
            elem.parent = alt
        t[0] = alt

    def p_rule_with_binding(self, t):
        """unit_with_binding : optional_binding_block unit"""
        unit = t[2]
        binding = t[1]
        if binding:
            compiled_binding = []
            for b in binding:
                cb = self.compile_binding(b)
                compiled_binding.append(cb)
            unit.pre_match_binding = compiled_binding + unit.pre_match_binding
        t[0] = unit

    def p_unit_term(self, t):
        'unit : term_sequence'
        sequence = t[1]
        if len(sequence) == 1:
            unit = sequence[0]
        else:
            unit = SequenceInputElement(tuple(sequence))
            for elem in sequence:
                elem.parent = unit
        t[0] = unit

    def p_term_sequence(self, t):
        """term_sequence : term_sequence term_with_binding"""
        sequence = t[1]
        term = t[2]
        sequence.append(term)
        t[0] = sequence

    def p_term_sequence_unit(self, t):
        """term_sequence : term_with_binding"""
        term = t[1]
        t[0] = [term]

    def compile_binding(self, binding):
        if binding is None:
            return None
        r, binding_code = self.binding_parser.parse(binding)
        #print("Binding parsed:", r, binding_code)
        if not r:
            print("Failed to parse binding code")
            print("Code:")
            print(binding)
            print(binding_code)
            raise Exception("Failed to parse binding code")
        return binding_code

    def p_term_with_binding(self, t):
        'term_with_binding : term optional_binding_block'
        term = t[1]
        binding = t[2]
        if binding:
            compiled_binding = []
            for b in binding:
                cb = self.compile_binding(b)
                compiled_binding.append(cb)
            term.post_match_binding = term.post_match_binding + compiled_binding

        t[0] = term

    def p_term_identifier(self, t):
        'term : identifier'
        identifier = t[1]
        grammar = self._grammar_spec.get_grammar(identifier)
        if grammar:
            grammar_ref = GrammarRefElement(grammar, [])
            t[0] = grammar_ref
            return

        tokendef = self._grammar_spec.get_tokendef(identifier)
        if tokendef:
            term = identifier
            elem = KeywordInputElement(term)
            t[0] = elem
            return

        elem = UnresolvedInputElement(identifier)
        self._grammar_spec.add_unresolved_element(elem)
        t[0] = elem

    def p_term_identifier_with_args(self, t):
        'term : identifier LBRACKET optional_argument_sequence RBRACKET'
        grammarref = t[1]
        arguments = t[3]

        if arguments is None:
            arguments = []

        grammar = self._grammar_spec.get_grammar(grammarref)
        if grammar:
            elem = GrammarRefElement(grammar, arguments)
        else:
            elem = UnresolvedInputElement(grammarref, arguments)
            self._grammar_spec.add_unresolved_element(elem)
        t[0] = elem

    def p_optional_argument_sequence(self, t):
        """optional_argument_sequence : argument_sequence
                                      | empty"""
        argument_list = t[1]
        if argument_list is None:
            argument_list = []
        t[0] = argument_list

    def p_optional_token_arguments(self, t):
        """optional_token_arguments : token_argument_sequence
                                    | empty"""
        argument_list = t[1]
        if argument_list is None:
            argument_list = []
        t[0] = argument_list

    def p_argument_sequence(self, t):
        """argument_sequence : argument_sequence COMMA argument
                             | argument"""
        arglist = t[1]
        if len(t) == 2:
            arglist = [arglist]
        else:
            arg = t[3]
            arglist.append(arg)
        t[0] = arglist

    def p_token_argument_sequence(self, t):
        """token_argument_sequence : token_argument_sequence COMMA token_argument
                                   | token_argument"""
        arglist = t[1]
        if len(t) == 2:
            arglist = [arglist]
        else:
            arg = t[3]
            arglist.append(arg)
        t[0] = arglist

    def p_argument(self, t):
        """argument : argument_types"""
        arg = t[1]
        t[0] = CliArgument(arg)

    def p_token_argument(self, t):
        """token_argument : basic_types"""
        arg = t[1]
        t[0] = CliTokenArgument(arg)

    def p_argument_with_param_name(self, t):
        """argument : parameter_name ASSIGN argument_types"""
        param = t[1]
        arg = t[3]
        t[0] = CliArgument(arg, param_name=param)

    def p_token_argument_with_param_name(self, t):
        """token_argument : token_parameter_name ASSIGN basic_types"""
        param = t[1]
        arg = t[3]
        t[0] = CliTokenArgument(arg, param_name=param)

    def p_argument_types(self, t):
        """argument_types : dollar_id
                          | basic_types"""
        arg = t[1]
        t[0] = arg

    def p_basic_types(self, t):
        """basic_types : number
                       | string_object
                       | boolean_true
                       | boolean_false
                       | none_object"""
        arg = t[1]
        t[0] = arg

    def p_string_object(self, t):
        """string_object : quoted_string"""
        quoted_string = t[1]
        actual_string = quoted_string[1:-1]
        string_object = convert_to_python_string(actual_string)
        t[0] = string_object

    def p_boolean_true(self, t):
        """boolean_true : TRUE"""
        t[0] = True

    def p_boolean_false(self, t):
        """boolean_false : FALSE"""
        t[0] = False

    def p_none_object(self, t):
        """none_object : NONE"""
        t[0] = None

    def p_term_quoted_string(self, t):
        'term : string_object'
        string_object = t[1]
        elem = ConstantInputElement(string_object)
        t[0] = elem

    def p_identifier(self, t):
        """identifier : IDENTIFIER
                      | usable_keywords"""
        identifier = t[1]
        t[0] = identifier

    def p_usable_keywords(self, t):
        """usable_keywords : IMPORT
                           | TOKEN
                           | TRUE
                           | FALSE
                           | NONE"""
        keyword = t[1]
        t[0] = keyword

    def p_optional_binding_block(self, t):
        """optional_binding_block : binding_block_sequence
                                  | empty"""
        binding = t[1] if t[1] else []
        t[0] = binding

    def p_binding_block_sequence(self, t):
        """binding_block_sequence : binding_block_sequence binding_block
                                  | binding_block"""
        if len(t) == 2:
            binding_block = t[1]
            binding_block_sequence = []
        else:
            binding_block = t[2]
            binding_block_sequence = t[1]

        if binding_block:
            binding_block_sequence.append(binding_block)
        t[0] = binding_block_sequence

    def p_binding_block(self, t):
        'binding_block : OPEN_BINDING binding_body CLOSE_BINDING'
        binding_text = t[2]
        t[0] = binding_text

    def p_binding_body(self, t):
        """binding_body : binding_text
                        | empty"""
        binding_text = t[1] if t[1] else ""
        t[0] = binding_text

    def p_binding_text(self, t):
        '''binding_text : binding_text binding_segment
                        | binding_segment'''

        if len(t) == 2:
            binding_segment = t[1]
            binding_text = binding_segment
        else:
            binding_text = t[1]
            binding_segment = t[2]
            binding_text += binding_segment

        t[0] = binding_text

    def p_binding_segment(self, t):
        """binding_segment : BOUND_CONTENT
                           | ESCAPED_CHAR"""
        binding_segment = t[1]
        t[0] = binding_segment

    def p_optional_parameter_list(self, t):
        """optional_parameter_list : LBRACKET parameter_list RBRACKET
                                   | LBRACKET empty RBRACKET
                                   | empty"""
        lbracket = t[1]
        if lbracket is None:
            t[0] =  []
        else:
            param_list = t[2]
            if param_list is None:
                t[0] =  []
            else:
                t[0] = param_list

    def p_parameter_list(self, t):
        """parameter_list : parameter_list COMMA parameter
                          | parameter"""
        if len(t) == 2:
            parameter = t[1]
            param_list = [parameter]
        else:
            parameter = t[3]
            param_list = t[1]
            param_list.append(parameter)
        t[0] = param_list

    def p_parameter(self, t):
        """parameter : parameter_name"""
        param = t[1]
        t[0] = CliParameter(param)

    def p_parameter_with_defvalue(self, t):
        """parameter : parameter_name ASSIGN parameter_value"""
        param = t[1]
        value = t[3]
        t[0] = CliParameter(param, defvalue=value, has_def_value=True)

    def p_parameter_name(self, t):
        """parameter_name : dollar_name"""
        param_id = t[1]
        t[0] = param_id

    def p_token_parameter_name(self, t):
        """token_parameter_name : identifier"""
        param_id = t[1]
        t[0] = param_id

    def p_parameter_value(self, t):
        """parameter_value : string_object
                           | dollar_name
                           | number"""
        parameter_value = t[1]
        t[0] = parameter_value

    def p_unused_token(self, t):
        """unused_token : eof
                        | AND
                        | NEWLINE
                        | ESCAPED_NEWLINE
                        | SINGLE_BACKSLASH
                        | QUOTED_INCOMPLETE_STR
                        | quoted_split_string"""
        print("Syntax error (Unused token)")
        self.p_error(t)

    def parse(self, input_str):
        self.lexer.enter_state(NessaidCliLexer.INITIAL_STATE)
        return self.parser.parse(input_str, lexer=self.lexer.lexer)

    def __init__(self, stdin=None, stdout=None, stderr=None):
        self._grammar_spec = GrammarSpecification()
        self._lexer = NessaidCliLexer(stdin=stdin, stdout=stdout, stderr=stderr)
        self.binding_parser = NessaidCliBindingParser(stdin=stdin, stdout=stdout, stderr=stderr)
        super().__init__(stdin=stdin, stdout=stdout, stderr=stderr)


def compile(input_str: str):
    """Compile the grammar specification in str format to a GrammarSpecification object

    :param input_str: The grammar specification as string
    :returns: a GrammarSpecification object which will contain the parsed grammars and token definitions
    :rtype: GrammarSpecification
    """

    parser = NessaidCliParser()
    output = parser.parse(input_str)
    return output


# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

from nessaid_cli.lex_yacc_common import NessaidCliLexerCommon, NessaidCliParserCommon

from nessaid_cli.binding_parser.binding_objects import (
    BindingCode,
    AssignmentStatement,
    BindingCall,
    FunctionCall,
    BindingStrObject
)


class NessaidCliBindingLexer(NessaidCliLexerCommon):

    tokens = NessaidCliLexerCommon.tokens + (
        'CALL',
        'TRUE',
        'FALSE',
        'NONE',
    )

    t_ignore_COMMENT = NessaidCliLexerCommon.common_COMMENT

    def t_TRUE(self, t):
        "True"
        return t

    def t_FALSE(self, t):
        "False"
        return t

    def t_NONE(self, t):
        "None"
        return t

    def t_CALL(self, t):
        "call"
        return t

    def __init__(self, filename="", starting_lineno=1, starting_linepos=0, stdin=None, stdout=None, stderr=None):
        # Build the lexer
        self._filename = filename
        self._starting_lineno = starting_lineno
        super().__init__(lineno=starting_lineno, linepos=starting_linepos, stdin=stdin, stdout=stdout, stderr=stderr)


class NessaidCliBindingParser(NessaidCliParserCommon):

    tokens = NessaidCliBindingLexer.tokens

    def p_binding_content(self, t):
        """binding_code : empty
                        | content"""
        binding_blocks = t[1]
        binding_code = BindingCode(binding_blocks) if binding_blocks else BindingCode([])
        t[0] = binding_code

    def p_content(self, t):
        """content : block
                   | content block"""
        if len(t) == 2:
            content = [t[1]]
        else:
            content = t[1]
            block = t[2]
            content.append(block)
        t[0] = content

    def p_block(self, t):
        """block : assign_block SEMICOLON
                 | call_block SEMICOLON
                 | function_block SEMICOLON
                 | unused_token"""
        t[0] = t[1]

    def p_assign_block(self, t):
        """assign_block : lhs_block ASSIGN rhs_block"""
        assign_block = AssignmentStatement(t[1], t[3])
        t[0] = assign_block

    def p_lhs_block(self, t):
        """lhs_block : dollar_name"""
        lhs_block = t[1]
        t[0] = lhs_block

    def p_rhs_block(self, t):
        """rhs_block : argument"""
        argument = t[1]
        t[0] = argument

    def p_call_block(self, t):
        """call_block : CALL identifier argument_block"""
        if len(t) == 4:
            call_block = BindingCall(t[2], t[3])
        else:
            call_block = BindingCall(t[2], [])
        t[0] = call_block

    def p_function_block(self, t):
        """function_block : identifier argument_block"""
        if len(t) == 2:
            function_block = FunctionCall(t[1], [])
        else:
            function_block = FunctionCall(t[1], t[2])
        t[0] = function_block

    def p_identifier(self, t):
        """identifier : IDENTIFIER
                      | usable_keywords"""
        identifier = t[1]
        t[0] = identifier

    def p_usable_keywords(self, t):
        """usable_keywords : CALL
                           | TRUE
                           | FALSE
                           | NONE"""
        keyword = t[1]
        t[0] = keyword

    def p_argument_block(self, t):
        """argument_block : LPAREN optional_argument_list RPAREN"""
        argument_list = t[2]
        t[0] = argument_list

    def p_optional_argument_list(self, t):
        """optional_argument_list : argument_list
                                  | empty"""
        argument_list = t[1]
        if argument_list is None:
            argument_list = []
        t[0] = argument_list

    def p_argument_list(self, t):
        """argument_list : argument_list COMMA argument
                         | argument"""
        arguments = t[1]
        if len(t) == 2:
            arguments = [arguments]
        else:
            argument = t[3]
            arguments.append(argument)
        t[0] = arguments

    def p_argument(self, t):
        """argument : dollar_id
                    | call_block
                    | function_block
                    | binding_object"""
        argument = t[1]
        t[0] = argument

    def p_binding_object(self, t):
        """binding_object : number
                          | string_object
                          | boolean_true
                          | boolean_false
                          | none_object"""
        binding_object = t[1]
        t[0] = binding_object

    def p_string_object(self, t):
        """string_object : quoted_string"""
        quoted_str = t[1][1:-1]
        t[0] = BindingStrObject(quoted_str)

    def p_boolean_true(self, t):
        """boolean_true : TRUE"""
        t[0] = True

    def p_boolean_false(self, t):
        """boolean_false : FALSE"""
        t[0] = False

    def p_none_object(self, t):
        """none_object : NONE"""
        t[0] = None

    def p_unused_token(self, t):
        """unused_token : ESCAPED_NEWLINE
                        | NEWLINE
                        | LBRACKET
                        | RBRACKET
                        | OR
                        | AND
                        | COLON
                        | LBRACE
                        | RBRACE
                        | MULTIPLY
                        | ESCAPED_CHAR
                        | SINGLE_BACKSLASH
                        | QUOTED_INCOMPLETE_STR
                        | quoted_split_string
                        | eof"""
        print("Syntax error (Unused token)")
        pass

    def __init__(self, filename="", starting_lineno=1, stdin=None, stdout=None, stderr=None):
        self._filename = filename
        self._starting_lineno = starting_lineno
        self._lexer = NessaidCliBindingLexer(filename=filename, starting_lineno=starting_lineno)
        super().__init__(stdin=stdin, stdout=stdout, stderr=stderr)

    def parse(self, input_str):
        try:
            self.lexer.enter_state(NessaidCliLexerCommon.INITIAL_STATE)
            code = self.parser.parse(input_str, lexer=self.lexer.lexer)
            if not isinstance(code, BindingCode):
                errmsg = "Expected {} got {}".format("BindingCode", code if not code else type(code))
                return False, errmsg
            return True, code
        except Exception as e:
            print("Error:", e)
            return False, str(e)


def parse(input_str):
    parser = NessaidCliBindingParser(filename="<SNIPPET>")
    parser.parse(input_str)



# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

from nessaid_cli.tokens import CliToken, MATCH_FAILURE
from nessaid_cli.lex_yacc_common import DollarVariable
from nessaid_cli.utils import ExtendedString


class CliParameter(ExtendedString):

    def __init__(self, value, defvalue=None, has_def_value=False):
        if type(has_def_value) is not bool:
            raise AttributeError("Attribute has_def_value should be Boolean")
        super().__init__(value, defvalue=defvalue, has_def_value=has_def_value)

    def __repr__(self):
        if not self.has_def_value:
            return super().__repr__()
        else:
            return "{}={}".format(str(self), self.defvalue)

    @property
    def name(self):
        return str(self)

class CliArgument():

    def __init__(self, value, param_name=None):
        if param_name is not None:
            if type(param_name) is not DollarVariable:
                raise AttributeError("Attribute param_name should be DollarVariable")
        self._value = value
        self._param_name = param_name

    @property
    def value(self):
        return self._value

    @property
    def param_name(self):
        return self._param_name

    def __repr__(self):
        if not self.param_name:
            return str(self.value)
        else:
            return "{}={}".format(str(self.param_name), str(self.value))


class CliTokenArgument(CliArgument):

    def __init__(self, value, param_name=None):
        if param_name is not None:
            if type(param_name) is not str:
                raise AttributeError("Attribute param_name should be str")
        self._value = value
        self._param_name = param_name


class TokenHierarchyElement():

    def __init__(self, element, path):
        self._element = element
        self._path = tuple(path)
        self._named_vars = {}
        self._numbered_vars = {}
        self._input_sequence = []

    @property
    def named_variables(self):
        return self._named_vars

    @property
    def token_variables(self):
        return self._numbered_vars

    @property
    def input_sequence(self):
        return self._input_sequence

    @property
    def element(self):
        return self._element

    def add_named_variable(self, var):
        self._named_vars[var.var_id] = var

    def add_numbered_variable(self, var):
        self._numbered_vars[var.var_id] = var

    @property
    def path(self):
        return self._path

    def __eq__(self, rhs):
        if isinstance(rhs, TokenHierarchyElement):
            return (self._element == rhs._element) and (self._path == rhs.path)
        return False

    def __hash__(self):
        return hash(self._element) + hash(self._path)


class TokenLookup():

    def __init__(self, token, root_grammar=None):
        self.token = token
        self.path = []
        self._root_grammar = root_grammar
        self._parent_stack = []

    @property
    def name(self):
        return self.token.value

    def add_prefix(self, prefix):
        self.path = [prefix] + self.path

    def get_element_hierarchy(self):
        self._parent_stack = []
        hierarchy = []
        level = 0
        parent_path = self.path.copy()
        while True:
            parent = self.get_parent(level)
            hierarchy.append(TokenHierarchyElement(parent, parent_path.copy()))
            parent_path.pop()
            level += 1
            if parent is self._root_grammar:
                self._parent_stack = []
                return hierarchy

    def __eq__(self, rhs):
        if isinstance(rhs, TokenLookup):
            if self._root_grammar is rhs._root_grammar:
                if (self.token == rhs.token) and (self.path == rhs.path):
                    return True
        return False

    def __hash__(self):
        return hash(self.token) + hash(tuple(self.path))

    def __repr__(self):
        return "{}:<{}>".format(self.token, ",".join(str(p) for p in self.path))

    def _locate(self, path):
        parent = self._root_grammar

        while path:
            position = path.pop(0)
            parent = parent.get(position)
        return parent

    def get_parent(self, parent_level):
        """
        R  P2  P1 T
        0  p2  p1 pt
        """
        if not self._parent_stack:
            level = 0
            parent = self.token

            self._parent_stack.append(parent)
            level += 1

            while parent:
                if type(parent) in [NamedGrammar]:
                    parent = self._locate(self.path[1:][:-level])
                else:
                    parent = parent.parent
                self._parent_stack.append(parent)
                if parent is self._root_grammar:
                    break
                level += 1

        return self._parent_stack[parent_level]


    def next(self):
        nexts = []
        self._parent_stack = []
        # import pdb; pdb.set_trace()
        level = 0
        parent = self.get_parent(level + 1)
        level += 1
        position = self.token.position + 1
        end_of_grammar = True

        while parent:
            elem = None
            if not isinstance(parent, AlternativeInputElement):
                while position < len(parent):
                    elem = parent.get(position)
                    if elem.repeat_count > 1:
                        firsts = elem.get(0).first(root_grammar=self._root_grammar)
                        for f in firsts:
                            f.add_prefix(elem.position)
                    else:
                        firsts = elem.first(root_grammar=self._root_grammar)

                    temp_parent = parent
                    temp_level = level

                    while temp_parent:
                        for f in firsts:
                            f.add_prefix(temp_parent.position)
                        if temp_parent is self._root_grammar:
                            break
                        temp_parent = self.get_parent(temp_level + 1)
                        temp_level += 1

                    nexts += firsts
                    if elem.mandatory:
                        end_of_grammar = False
                        break
                    position += 1

                    if parent.repeat_count > 1:
                        break
                if elem and elem.mandatory:
                    break

            if parent is self._root_grammar:
                break
            position = parent.position + 1
            parent = self.get_parent(level + 1)
            level += 1

        if end_of_grammar or not nexts:
            nexts.append(EndOfInpuToken())

        self._parent_stack = []
        return nexts


class _EndOfInpuToken(TokenLookup, CliToken):

    __instance = None

    @property
    def completable(self):
        return False

    def get_value(self, match_string=None):
        return None

    @staticmethod
    def getInstance():
        if _EndOfInpuToken.__instance == None:
            _EndOfInpuToken()
        return _EndOfInpuToken.__instance

    def __init__(self):
        if _EndOfInpuToken.__instance != None:
            raise Exception("This class is a singleton!")
        self._name = "$END_OF_INPUT"
        _EndOfInpuToken.__instance = self

    def __repr__(self):
        return "< End of Input >"

    def __str__(self):
        return self.__repr__()

    def __eq__(self, rhs):
        if isinstance(rhs, _EndOfInpuToken):
            return True
        return False

    def __hash__(self):
        return hash(self._name)

    def __not__(self):
        return True

    def truth(self):
        return False

    def is_(self, rhs):
        if isinstance(rhs, _EndOfInpuToken):
            return True
        return False

    def is_not(self, rhs):
        if isinstance(rhs, _EndOfInpuToken):
            return False
        return True

    def __nonzero__(self):
        return False

    def __bool__(self):
        return False

    def complete(self, *args, **kwargs):
        return []

    def match(self, token_input):
        return MATCH_FAILURE

    @property
    def helpstring(self):
        return str(self)

    @property
    def name(self):
        return self._name

    def next(self):
        return [self]


def EndOfInpuToken():
    return _EndOfInpuToken.getInstance()


class DuplicateDefinitionException(Exception):

    def __init__(self, name):
        self._name = name

    def __repr__(self):
        return 'DuplicateDefinitionException("{}")'.format(self._name)

    def __str__(self):
        return 'DuplicateDefinitionException("{}")'.format(self._name)


class DuplicateTokendefException(Exception):

    def __init__(self, name):
        self._name = name

    def __repr__(self):
        return 'DuplicateTokendefException("{}")'.format(self._name)

    def __str__(self):
        return 'DuplicateTokendefException("{}")'.format(self._name)


class ArgumentError(Exception):
    pass


class InputElement():

    def __init__(self):
        self._parent = None
        self._value = None
        self._pre_match_binding = []
        self._post_match_binding = []
        self._position = None
        self._repeat_count = 1

    def copy_extras(self, cp):
        cp._pre_match_binding = self._pre_match_binding
        cp._post_match_binding = self._post_match_binding
        cp.repeat_count = self.repeat_count

    def __str__(self):
        return repr(self)

    def __len__(self):
        return 1

    @property
    def repeat_count(self):
        return self._repeat_count

    @repeat_count.setter
    def repeat_count(self, value):
        if isinstance(value, int) and value > 0:
            self._repeat_count = value
        else:
            raise ValueError("Repeater should be > 0")

    @property
    def value(self):
        return self._value

    @property
    def parent(self):
        try:
            return self._parent
        except:
            return None

    @parent.setter
    def parent(self, p):
        if not isinstance(p, InputElementHolder):
            raise ValueError("Expected InputElementCollection object for parent")
        self._parent = p

    @property
    def position(self):
        if self._position is None:
            if type(self.parent) in [NamedGrammar, GrammarRefElement]:
                self._position = 0
            else:
                self._position = self.parent.index(self)
        return self._position

    @position.setter
    def position(self, pos):
        pos = int(pos)
        if pos < 0:
            raise ValueError("Position should be 0 or a positive integer")
        self._position = pos

    def validate_binding(self, b):
        if not isinstance(b, list):
            for e in b:
                if not isinstance(e, str):
                    raise ValueError("Expected str object for binding code")
            raise ValueError("Expected list object for binding code segments")
        return b

    @property
    def pre_match_binding(self):
        return self._pre_match_binding

    @pre_match_binding.setter
    def pre_match_binding(self, b):
        self._pre_match_binding = self.validate_binding(b)

    @property
    def post_match_binding(self):
        return self._post_match_binding

    @post_match_binding.setter
    def post_match_binding(self, b):
        self._post_match_binding = self.validate_binding(b)

    @property
    def mandatory(self):
        return False


class ConstantInputElement(InputElement):

    def __init__(self, keyword):
        super().__init__()
        self._value = keyword

    def copy(self):
        cp = ConstantInputElement(self._value)
        self.copy_extras(cp)
        return cp

    def __repr__(self, verbose=True):
        return self._value

    @property
    def mandatory(self):
        return True

    def first(self, root_grammar=None):
        first =  TokenLookup(self, root_grammar=root_grammar)
        if self.parent:
            first.add_prefix(self.position)
        return [first]


class KeywordInputElement(InputElement):

    def __init__(self, keyword):
        super().__init__()
        if isinstance(keyword, str):
            self._value = keyword
        else:
            raise ValueError("Expected str object")

    def copy(self):
        cp = KeywordInputElement(self._value)
        self.copy_extras(cp)
        return cp

    def __repr__(self, verbose=True):
        return self._value

    def __hash__(self):
        return self._value.__hash__()

    @property
    def mandatory(self):
        return True

    def first(self, root_grammar=None):
        first = TokenLookup(self, root_grammar=root_grammar)
        if self.parent:
            first.add_prefix(self.position)
        return [first]


class InputElementHolder(InputElement):
    pass


class InputElementCollection(InputElementHolder):

    def __init__(self, sequence):
        super().__init__()
        self._mandatory = None
        if isinstance(sequence, tuple):
            self._value = sequence
        else:
            raise ValueError("Expected tuple object")

    def __len__(self):
        return len(self._value)

    def copy(self):
        val = tuple([v.copy() for v in self._value])
        cp = self.__class__(val)
        for v in val:
            v.parent = cp
        self.copy_extras(cp)
        return cp

    def get(self, position):
        return self._value[position]

    def index(self, child):
        pos = 0
        for c in self._value:
            if id(c) == id(child):
                return pos
            pos += 1
        raise IndexError("{} doesnt have {}".format(self, child))

    def first(self, root_grammar=None):
        firsts = []
        for i in range(len(self)):
            elem = self.get(i)
            firsts += elem.first(root_grammar=root_grammar)
            if elem.mandatory:
                break

        if self.parent:
            for first in firsts:
                first.add_prefix(self.position)
        return firsts


class SequenceInputElement(InputElementCollection):

    def __repr__(self, verbose=True):
        _repr = "Seq" if verbose else ""
        if self.repeat_count > 1:
            _repeat = "*{}".format(self.repeat_count)
        else:
            _repeat = ""
        return _repr + '(' + ", ".join([e.__repr__(verbose=False) for e in self.value]) + ')' + _repeat

    def __len__(self):
        if self.repeat_count > 1:
            assert len(self._value) == 1
            return self.repeat_count
        else:
            return len(self._value)

    def get(self, position):
        if position > 0 and self.repeat_count > 1 and position < self.repeat_count:
            assert len(self._value) == 1
            elem = self._value[0].copy()
            elem.parent = self
            elem._position = position
            return elem
        return self._value[position]

    def index(self, child):
        if self.repeat_count > 1 and child.position < self.repeat_count:
            if child.parent == self:
                if child.position < len(self):
                    return child.position
        pos = 0
        for c in self._value:
            if id(c) == id(child):
                return pos
            pos += 1
        raise IndexError("{} doesnt have {}".format(self, child))

    @property
    def mandatory(self):
        if self._mandatory is not None:
            return self._mandatory
        for elem in self._value:
            if elem.mandatory:
                self._mandatory = True
                return True
        self._mandatory = False
        return False


class OptionalInputElement(InputElementCollection):

    def __repr__(self, verbose=True):
        _repr = "Opt" if verbose else ""
        return _repr + '{' + ", ".join([e.__repr__(verbose=False) for e in self.value]) + '}'


class AlternativeInputElement(InputElementCollection):

    def __repr__(self, verbose=True):
        _repr = "Alt" if verbose else ""
        return _repr + '(' + " | ".join([e.__repr__(verbose=False) for e in self.value]) + ')'

    @property
    def mandatory(self):
        if self._mandatory is not None:
            return self._mandatory
        for elem in self._value:
            if not elem.mandatory:
                self._mandatory = False
                return False
        self._mandatory = True
        return True

    def first(self, root_grammar=None):
        firsts = []
        for i in self._value:
            firsts += i.first(root_grammar=root_grammar)
        if self.parent:
            for first in firsts:
                first.add_prefix(self.position)
        return firsts


class NamedGrammar(InputElementHolder):

    def __init__(self, name, param_list, value):
        super().__init__()
        self._name = name
        self._param_list = []

        if isinstance(value, InputElement):
            self._value = value
        else:
            raise ValueError("Expected InputElement object")

        if not param_list:
            param_list = []

        for param in param_list:
            if not isinstance(param, str):
                raise ValueError("Parameter name should be str not {}".format(type(param)))
            if param in self._param_list:
                raise ValueError("Duplicate parameter name: {}".format(param))
            self._param_list.append(param)

    def copy(self):
        cp = NamedGrammar(self._name, self._param_list, self._value.copy())
        self.copy_extras(cp)
        return cp

    @property
    def param_list(self):
        return self._param_list

    @property
    def value(self):
        return self._value

    def get(self, position):
        if position == 0:
            return self._value
        raise IndexError("NamedGrammar: Invalid index: {}".format(position))

    def index(self, child):
        if id(self._value) == id(child):
            return 0
        raise IndexError("{} doesnt have {}".format(self, child))

    @property
    def parent(self):
        return None

    @parent.setter
    def parent(self, p):
        raise ValueError("Cannot set parent for named grammar")

    @property
    def position(self):
        return 0

    @property
    def name(self):
        return self._name

    @property
    def mandatory(self):
        return self._value.mandatory

    def first(self, root_grammar=None):
        firsts = self._value.first(root_grammar=root_grammar)
        for first in firsts:
            first.add_prefix(self.position)
        return firsts


def map_grammar_arguments(grammar_name, parameter_list, arglist):
    arg_count = len(arglist)
    param_list = parameter_list.copy()
    param_count = len(param_list)

    if param_count < arg_count:
        raise ArgumentError(
            "Grammar {} referenced with {} arguments. It has only {} parameters".format(
                grammar_name, arg_count, param_count))

    has_kwarg = False
    param_map = {}

    for arg in arglist:
        if arg.param_name:
            if arg.param_name in param_map:
                raise ArgumentError("Param {} got more than one argument".format(arg.param_name))
            has_kwarg = True
            for param in param_list:
                if arg.param_name == param:
                    param_map[arg.param_name] = arg
                    param_list.remove(param)
                    break
            else:
                raise ArgumentError("Could not match param {} for keyword argument".format(arg.param_name))
        else:
            if has_kwarg:
                raise ArgumentError("Positional argument should not follow keyword argument")
            param = param_list.pop(0)
            param_map[str(param)] = arg

    while param_list:
        param = param_list.pop(0)
        if param.has_def_value:
            param_map[str(param)] = CliArgument(param.defvalue)
        else:
            param_map[str(param)] = CliArgument(None)

    return param_map

class GrammarRefElement(InputElement):

    def __init__(self, grammar, arglist):
        super().__init__()
        if isinstance(grammar, NamedGrammar):
            self._value = grammar
            self._arglist = arglist if arglist else []
            if self._arglist:
                self._param_mapping = map_grammar_arguments(self.name, self._value.param_list, self._arglist)
        else:
            raise ValueError("Expected NamedGrammar object")

    def copy(self):
        cp = GrammarRefElement(self._value, self._arglist)
        self.copy_extras(cp)
        return cp

    @property
    def name(self):
        return self._value.name

    @property
    def value(self):
        return self._value

    @property
    def arg_list(self):
        return self._arglist

    @property
    def param_mapping(self):
        return self._param_mapping

    def get(self, position):
        if position == 0:
            return self._value
        raise IndexError("GrammarRefElement: Invalid index: {}".format(position))

    @property
    def mandatory(self):
        return self._value.mandatory

    def first(self, root_grammar=None):
        firsts = self._value.first(root_grammar=root_grammar)
        if self.parent:
            for first in firsts:
                first.add_prefix(self.position)
        return firsts


class UnresolvedInputElement(InputElement):

    def __init__(self, keyword, arglist=None):
        super().__init__()
        if isinstance(keyword, str):
            self._value = keyword
        else:
            raise ValueError("Expected str object")
        self._arglist = arglist if arglist else []

    @property
    def value(self):
        return self._value

    @property
    def arg_list(self):
        return self._arglist

    def copy(self):
        cp = UnresolvedInputElement(self._value)
        self.copy_extras(cp)
        return cp


class TokenClassDef():

    def __init__(self, classname, arglist):
        self._classname = classname
        self._arglist = [arg.value for arg in arglist]

    @property
    def classname(self):
        return self._classname

    @property
    def arglist(self):
        return self._arglist


class GrammarSpecification:

    def __init__(self):
        self._grammars = []
        self._tokens = []
        self._token_defs = {}
        self._named_grammars = {}
        self._unresolved_tokens = {}

    @property
    def grammars(self):
        return self._grammars

    @property
    def tokens(self):
        return self._tokens

    @property
    def token_defs(self):
        return self._token_defs

    @property
    def named_grammars(self):
        return self._named_grammars

    def add_grammar(self, grammar):
        if grammar.name in self._named_grammars:
            raise DuplicateDefinitionException(grammar.name)
        if grammar.name in self._token_defs:
            raise DuplicateDefinitionException(grammar.name)
        self._named_grammars[grammar.name] = grammar
        self._grammars.append(grammar)

    def add_token_def(self, tokenname, classdef):
        if tokenname in self._token_defs:
            raise DuplicateTokendefException(tokenname)
        if tokenname in self._named_grammars:
            raise DuplicateDefinitionException(tokenname)
        self._token_defs[tokenname] = classdef
        self._tokens.append(classdef)

    def add_unresolved_element(self, element):
        if element.value not in self._unresolved_tokens:
            self._unresolved_tokens[element.value] = []
        self._unresolved_tokens[element.value].append(element)

    def get_grammar(self, name):
        if name in self._named_grammars:
            return self._named_grammars[name]
        return None

    def get_tokendef(self, name):
        if name in self.token_defs:
            return self.token_defs[name]
        return None


# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import traceback

import ply.lex as lex
import ply.yacc as yacc

from nessaid_cli.utils import StdStreamsHolder, ExtendedString


class TokenizerIllegalCharError(Exception):

    def __init__(self, msg, token=None):
        self.token = token
        super().__init__(msg)


class TokenizerSyntaxError(Exception):

    def __init__(self, msg, token=None):
        self.token = token
        super().__init__(msg)


class TokenizerException(Exception):

    def __init__(self, msg, token=None):
        self.token = token
        super().__init__(msg)


class TokenInputString(ExtendedString):

    def __init__(self, value, lexpos, lexlen, quoted=False, quote_incomplete=False):
        if quote_incomplete is True:
            quoted = True
        super().__init__(value, lexpos=lexpos, lexlen=lexlen, quoted=quoted, quote_incomplete=quote_incomplete)


class NessaidCliTokenizerLexer(StdStreamsHolder):

    tokens = (
        'TEXT',
        'QUOTED_STR',
        'QUOTED_INCOMPLETE_STR',
    )

    def t_TEXT(self, t):
        r'([^"\n\\ \t])+'
        t.value = TokenInputString(t.value, t.lexpos, len(t.value))
        return t

    def t_QUOTED_STR(self, t):
        r'(")(([^"\n\r\t\\\0])|((\\\\)|(\\)(0|n|r|t|b|v|a|")))*(")'
        t.value = TokenInputString(t.value, t.lexpos, len(t.value), quoted=True)
        return t

    def t_QUOTED_INCOMPLETE_STR(self, t):
        r'(")(([^"\n\r\t\\\0])|((\\\\)|(\\)(0|n|r|t|b|v|a|")))*(\\)?'
        t.value = TokenInputString(t.value, t.lexpos, len(t.value), quote_incomplete=True)
        return t

    @property
    def lexer(self):
        return self._lexer

    def __init__(self, stdin=None, stdout=None, stderr=None):
        self._lexer = None
        self.init_streams(stdin=stdin, stdout=stdout, stderr=stderr)
        self._lexer = lex.lex(module=self)

    t_ignore  = ' \t'

    def t_error(self, t):
        raise TokenizerIllegalCharError("Illegal character: '{}'".format(t.value[0]), token=t.value[0])


class NessaidCliTokenizer(StdStreamsHolder):

    tokens = NessaidCliTokenizerLexer.tokens

    @property
    def lexer(self):
        return self._lexer

    @property
    def parser(self):
        return self._parser

    def __init__(self, stdin=None, stdout=None, stderr=None):
        self._lexer = None
        self._parser = None
        self.init_streams(stdin=stdin, stdout=stdout, stderr=stderr)

        self._lexer = NessaidCliTokenizerLexer(stdin=stdin, stdout=stdout, stderr=stderr)
        self._parser = yacc.yacc(module=self, debug=False, write_tables=False)

    def parse(self, input_str):
        try:
            return self.parser.parse(input_str, lexer=self.lexer.lexer)
        except TokenizerIllegalCharError as e:
            error_msg = "Illegal character error while tokenizing." + " Token: {}".format(e.token) if e.token else ""
            raise TokenizerException(error_msg, token=e.token)
        except TokenizerSyntaxError as e:
            error_msg = "Syntax error while tokenizing." + " Token: {}".format(e.token) if e.token else ""
            raise TokenizerException(error_msg, token=e.token)
        except Exception as e:
            error_msg = "Exception while tokenizing: {}".format(str(e)) + " Token: {}".format(e.token) if e.token else ""
            return e

    def p_line_content(self, t):
        """line_content : line empty"""
        line = t[1]
        t[0] = line

    def p_line(self, t):
        """line : line segment
                | empty"""
        line = t[1]
        if line is None:
            line = []
        else:
            segment = t[2]
            line.append(segment)
        t[0] = line

    def p_segment(self, t):
        """segment : text
                   | quoted_string
                   | incomplete_quoted_string"""
        segment = t[1]
        t[0] = segment

    def p_quoted_string(self, t):
        'quoted_string : QUOTED_STR'
        quoted_string = t[1]
        t[0] = quoted_string

    def p_incomplete_quoted_string(self, t):
        'incomplete_quoted_string : QUOTED_INCOMPLETE_STR'
        quoted_string = t[1]
        t[0] = quoted_string

    def p_text(self, t):
        'text : TEXT'
        text = t[1]
        t[0] = text

    def p_empty(self, t):
        'empty :'
        pass

    def p_error(self, t):
        raise Exception("Error parsing input line")


def tokenize(input_str):
    parser = NessaidCliTokenizer()
    tokens = []
    try:
        tokens = parser.parse(input_str)
    except TokenizerException as e:
        parser.error("Failed to tokenize input:", e)
    except Exception as e:
        parser.error("Exception tokenizing input: {}\nException: {}: {}".format(input_str, type(e), e))
    return tokens


def test_main():
    while True:
        try:
            line = input("Test: ")
            tokens = tokenize(line)
            print("Line:", line)
            print("Tokens:", tokens)
            print("Tokens Inputs:", [t.value for t in tokens])
        except KeyboardInterrupt:
            break
        except Exception as e:
            print("Exception:", type(e), e)


if __name__ == '__main__':
    test_main()


# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import inspect
import textwrap

from nessaid_cli.cli import NessaidCli
from nessaid_cli.compiler import compile


class NessaidCmd(NessaidCli):
    """
    """
    """
    Base Cmd class

    The class can be initialized with cli_hook_prefix argument whose default value is do_
    All the functions starting with the prefix will be treated as Cmd command handlers
    The input will be matched as per the grammar specification given as the docstring and
    the function. The global grammar definitions should go as the derived class's docstring
    """

    def generate_grammar_name(self, hookmethod):
        """Generate the grammar name used in the grammar

        Returns the grammar name autogenerated from the Cmd functions.

        :param hookmethod: The method name with which the associated grammar is used
        :returns: The name of the generated grammar block
        :rtype: str
        """

        return hookmethod.__name__

    def generate_root_grammar_name(self):
        """Generate the root grammar name used in the Cmd instance

        Returns the root grammar name autogenerated from the Cmd functions.

        :returns: The name of the root grammar specification used in the Cmd instance
        :rtype: str
        """

        return self.__class__.__name__.split(".")[0]

    def format_grammar(self, grammar_text):
        formatted = ""
        if grammar_text:
            grammar_text = grammar_text.replace("\r\n", "\n")
            grammar_text = grammar_text.replace("\r", "\n")
            formatted = textwrap.dedent(grammar_text)
        return formatted

    def __init__(self, loop=None, prompt=None, cli_hook_prefix="do_", cli_nargs=3,
                 stdin=None, stdout=None, stderr=None, completekey='tab', use_rawinput=True, show_grammar=False):
        """Creates a Cmd instance

        :param loop: the event loop used to run the Cmd loop.
        :param prompt: The prombp string for the Cmd instance
        :param cli_hook_prefix: The prefix of the methods in the class to be converted to Cmd commands
        :param cli_nargs: Number of arguments the generated Cmd handlers should have.
        :param show_grammar: print the generated grammar before the Cmd prompt.
        """

        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr

        self.execute_line = self.exec_line
        self.execute_args = self.exec_args
        grammar_text = self.__doc__ if self.__doc__ else ""
        grammar_hooks = [getattr(self, f) for f in dir(self) if f.startswith(cli_hook_prefix) and callable(getattr(self, f))]
        grammar_alternatives = []

        for hook in grammar_hooks:
            if hook.__doc__:
                argspec = inspect.getfullargspec(hook)
                argnames = argspec.args[1:]
                gramar_name = self.generate_grammar_name(hook)
                grammar_alternatives.append(gramar_name)
                argstring = ", ".join(["$arg_" + str(n + 1) for n in range(cli_nargs)])
                grammar_name_line = "\n\n    {grammar_name}[{argstring}]:\n".format(
                    grammar_name=gramar_name, argstring=argstring)
                if argnames:
                    grammar_name_line += "      <<\n"
                    for arg in argnames:
                        grammar_name_line += "        ${argname} = \"\";\n".format(argname=arg)
                    grammar_name_line += "      >>\n"
                hook_grammar = grammar_name_line
                hook_grammar +=  "      ("
                hook_grammar += hook.__doc__.rstrip() + "\n"
                hook_grammar +=  "      )\n"
                hook_grammar +=  "      <<call {grammar_name}({grammar_args});>>".format(
                    grammar_name=self.get_cli_hook(gramar_name),
                    grammar_args=", ".join(["$" + arg for arg in argnames]))
                hook_grammar += "\n      ;"

                grammar_text += hook_grammar

        if grammar_alternatives:
            root_grammar_name = self.generate_root_grammar_name()
            root_grammar = "\n\n    {grammar_name}[{argstring}]:\n".format(grammar_name=root_grammar_name, argstring=argstring)
            root_grammar += "      " + "\n      |\n      ".join(grammar_alternatives)
            root_grammar += "\n      ;\n"

            grammar_text += root_grammar
            grammar_text = self.format_grammar(grammar_text)

            if show_grammar:
                self.print("# Generated CLI grammar:")
                self.print(grammar_text)

            grammar_set = compile(grammar_text)
            super().__init__(grammar_set, prompt=prompt, loop=loop,
                 stdin=stdin, stdout=stdout, stderr=stderr, completekey=completekey, use_rawinput=use_rawinput)

    def do__exit(self):
        """
        "exit" | "quit" | "end"
        """
        self.exit_loop()

    async def cmdloop(self, intro=None):
        """Runs the cmd loop until exit

        :returns: None
        :rtype: None
        """

        return await super().cmdloop(grammarname=self.generate_root_grammar_name(), intro=intro)

    @classmethod
    async def execute_args(cls, *args):
        cmd = cls(prompt="# ", show_grammar=False)
        return await cmd.exec_args(*args)

    @classmethod
    async def execute_line(cls, line):
        cmd = cls(prompt="# ", show_grammar=False)
        return await cmd.exec_line(line)

    async def exec_args(self, *args):
        modified_args = []
        chars_to_check = [" "]
        for arg in args:
            if not (arg.startswith('"') and arg.endswith('"')):
                need_quoting = False
                for c in chars_to_check:
                    if arg.strip():
                        if c in arg:
                            need_quoting = True
                            break
                if need_quoting:
                    modified_args.append('"' + arg + '"')
                    continue
            modified_args.append(arg)

        args = modified_args
        line = " ".join(args)
        return await self.exec_line(line)


    async def exec_line(self, line):
        try:
            grammar = self.generate_root_grammar_name()
        except Exception as e:
            self.error("Exception getting root grammar:")
            return 1
        try:
            self.enter_grammar(self.generate_root_grammar_name())
        except Exception as e:
            self.error("Exception entering grammar:", grammar, "Error:", e)
            return 2
        try:
            await self.cli_exec_init()
            await super().exec_line(line)
        except Exception as e:
            self.error("Exception executing grammar:", grammar, "input:", line, "Error:", e)
            return 3
        finally:
            self.exit_grammar()
        return 0




# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import ply.lex as lex
import ply.yacc as yacc

from nessaid_cli.utils import StdStreamsHolder, ExtendedString


class CliLexerError(Exception):
    pass

class CliSyntaxError(Exception):
    pass


class DollarNumber(ExtendedString):

    def __init__(self, value):
        super().__init__(value)


class DollarVariable(ExtendedString):

    def __init__(self, value):
        super().__init__(value)


class NessaidCliLexerCommon(StdStreamsHolder):

    INITIAL_STATE = 'INITIAL'

    tokens = (
        'LPAREN',
        'RPAREN',
        'LBRACE',
        'RBRACE',
        'LBRACKET',
        'RBRACKET',
        'OR',
        'AND',
        'COMMA',
        'COLON',
        'MULTIPLY',
        'SEMICOLON',
        'ASSIGN',
        'IDENTIFIER',
        'INTEGER',
        'FLOAT',
        'DOLLAR_NUMBER_ID',
        'DOLLAR_VAR_ID',
        'NEWLINE',
        'ESCAPED_CHAR',
        'SINGLE_BACKSLASH',
        'ESCAPED_NEWLINE',
        'QUOTED_STR',
        'QUOTED_INCOMPLETE_STR',
        'HEX_NUMBER',
        'eof',
    )

    t_LPAREN      = r'\('
    t_RPAREN      = r'\)'
    t_LBRACE      = r'\{'
    t_RBRACE      = r'\}'
    t_LBRACKET    = r'\['
    t_RBRACKET    = r'\]'
    t_COLON       = r':'
    t_SEMICOLON   = r';'

    t_OR          = r'\|'
    t_AND         = r'\&'
    t_COMMA       = r','
    t_ASSIGN      = r'='
    t_MULTIPLY    = r'\*'

    t_IDENTIFIER  = r'[A-Za-z_][-a-zA-Z0-9_]*'
    t_INTEGER     = r'[\+-]?(([0])|(([1-9])([0-9]*)))'
    t_FLOAT       = r'[\+-]?(([0]?)|(([1-9])([0-9]*)))(\.)([0-9]+)'
    t_QUOTED_STR  = r'(")(([^"\n\r\t\\\0])|((\\\\)|(\\)(0|n|r|t|b|v|a|")))*(")'
    t_HEX_NUMBER  = r'(0)(x|X)([a-fA-F0-9]+)'

    t_DOLLAR_NUMBER_ID   = r'[\$][0-9]+'
    t_DOLLAR_VAR_ID      = r'[\$][A-Za-z_][a-zA-Z0-9_]*'

    def t_eof(self, t):
        return None

    t_ignore  = ' \t'

    def t_NEWLINE(self, t):
        r'(\n|\r\n|\r)+'
        self.update_counters(t)

    def t_ESCAPED_CHAR(self, t):
        r'\\([^\n\r])'
        return t

    def t_ESCAPED_NEWLINE(self, t):
        r'\\(\n|\r\n|\r)'
        self.update_counters(t)

    def t_error(self, t):
        self.update_counters(t)

        def _repr(t):
            v = t
            if v:
                v.replace('\n', "'\\n'")
                v.replace('\t', "'\\t'")
            return v

        err_msg = "Line: {line} Postion: {position} State: {state}: Illegal character {char}".format(
            line=self.lineno,
            position=self.linepos,
            state=self.state,
            char=_repr(t.value[0])
        )
        self.error(err_msg)
        raise CliLexerError(err_msg)

    def count_newlines(self, pattern, **kwargs):
        m = pattern
        m.replace("\r\n", "\n")
        m.replace("\r", "\n")
        return m.count("\n")

    @property
    def state(self):
        if self._lex_states:
            return self._lex_states[-1]
        return NessaidCliLexerCommon.INITIAL_STATE

    def enter_state(self, state):
        self._lex_states.append(state)
        self.lexer.begin(state)

    def exit_state(self, state):
        curstate = self._lex_states.pop()
        assert curstate == state
        self.lexer.begin(self._lex_states[-1] if self._lex_states else NessaidCliLexerCommon.INITIAL_STATE)

    def common_COMMENT(self, t):
        r'(/\*(.|\n)*?\*/)|(//.*)|(\#.*)'
        self.update_counters(t)

    @property
    def lexer(self):
        return self._lexer

    @property
    def lineno(self):
        return self._lineno

    @lineno.setter
    def lineno(self, n):
        _incr = n - self._lineno
        self._lineno = int(n)
        self.lexer.lineno += _incr

    @property
    def linepos(self):
        try:
            if self.lexer:
                return self.lexer.lexpos - self._size_till_last_newline
        except Exception:
            pass
        return 0

    def update_counters(self, t):
        count = self.count_newlines(t.value)
        self.lineno += count
        if count:
            self._size_till_last_newline = t.lexpos + max(t.value.rfind("\r"), t.value.rfind("\n")) + 1

    @property
    def size_till_last_newline(self):
        return self._size_till_last_newline

    def __init__(self, lineno=1, linepos=0, stdin=None, stdout=None, stderr=None):

        self.init_streams(stdin=stdin, stdout=stdout, stderr=stderr)

        self._lexer = None
        self._lineno = lineno
        self._size_till_last_newline = 0
        self._lex_states = []
        self._lexer = lex.lex(module=self)


class NessaidCliParserCommon(StdStreamsHolder):

    def __init__(self, stdin=None, stdout=None, stderr=None):

        self.init_streams(stdin=stdin, stdout=stdout, stderr=stderr)

        self._parser = None
        self._parser = yacc.yacc(module=self, debug=False, write_tables=False)

    @property
    def lexer(self):
        return self._lexer

    @property
    def parser(self):
        return self._parser

    def p_empty(self, t):
        'empty :'
        pass

    def p_quoted_string(self, t):
        """quoted_string : QUOTED_STR"""
        quoted_string = t[1]
        t[0] = quoted_string

    def p_quoted_split_string(self, t):
        """quoted_split_string : quoted_split_string QUOTED_STR
                               | QUOTED_STR"""
        quoted_string = t[1]
        if len(t) == 3:
            quoted_string_to_add = t[2]
            quoted_string = quoted_string + quoted_string_to_add

        t[0] = quoted_string

    def p_number(self, t):
        """number : integer
                  | float"""
        number = t[1]
        t[0] = number

    def p_integer(self, t):
        """integer : INTEGER
                   | HEX_NUMBER"""
        int_str =  t[1]
        integer = int(int_str)
        t[0] = integer

    def p_float(self, t):
        """float : FLOAT"""
        float_str = t[1]
        float_val = float(float_str)
        t[0] = float_val

    def p_dollar_id(self, t):
        """dollar_id : dollar_name
                     | dollar_number"""
        dollar_id = t[1]
        t[0] = dollar_id

    def p_dollar_name(self, t):
        'dollar_name : DOLLAR_VAR_ID'
        dollar_name = t[1]
        t[0] = DollarVariable(dollar_name)

    def p_dollar_number(self, t):
        'dollar_number : DOLLAR_NUMBER_ID'
        dollar_number = t[1]
        t[0] = DollarNumber(dollar_number)

    def p_error(self, t):

        def _repr(t):
            v = t.value if t is not None and hasattr(t, 'value') else t
            if v:
                v.replace('\n', "'\\n'")
                v.replace('\t', "'\\t'")
            return v

        err_msg = "Line: {line} Postion: {position} State: {state}: Syntax Error at: {token}".format(
            line=self.lexer.lineno,
            position=self.lexer.linepos,
            state=self.lexer.state,
            token=_repr(t))

        self.error(err_msg)
        raise CliSyntaxError(err_msg)


# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import sys
import asyncio

from nessaid_cli.utils import StdStreamsHolder, convert_to_python_string


from nessaid_cli.tokens import (
    CliToken,
    MATCH_SUCCESS,
    MATCH_PARTIAL,
    MATCH_FAILURE,
    MATCH_AMBIGUOUS
)

from nessaid_cli.elements import (
    EndOfInpuToken,
    NamedGrammar,
    GrammarRefElement,
    AlternativeInputElement,
    GrammarSpecification,
    map_grammar_arguments,
    TokenHierarchyElement
)

from nessaid_cli.elements import (
    SequenceInputElement,
    OptionalInputElement,
    CliArgument
)


from nessaid_cli.binding_parser.binding_objects import (
    BindingCall,
    FunctionCall,
    NamedVariable,
    TokenVariable,
    BindingVariable,
    AssignmentStatement,
)

from nessaid_cli.lex_yacc_common import (
    DollarNumber,
    DollarVariable
)


class ParsingResult():

    def __init__(self):
        self.result = None
        self.error = None
        self.matched_sequence = []
        self.last_token = {}
        self.offending_token = None
        self.offending_token_position = None
        self.next_tokens = []
        self.next_constant_token = None
        self.matched_values = []
        self.case_insensitive = False

    def as_dict(self):
        return {
            'result': self.result,
            'error': self.error,
            'matched_sequence': self.matched_sequence,
            'matched_values': self.matched_values,
            'offending_token': self.offending_token,
            'offending_token_position': self.offending_token_position,
            'next_tokens': self.next_tokens,
            'last_token': self.last_token,
            'next_constant_token': self.next_constant_token,
            'case_insensitive': self.case_insensitive,
        }

    def __repr__(self):
        return str(self.as_dict())

    def __str__(self):
        return str(self.as_dict())

    def common_prefix(self, l):
        l.sort(reverse = False)
        if not l:
            return ""
        elif len(l) == 1:
            return l[0]
        else:
            str1 = l[0]
            str2 = l[-1]

        n1 = len(str1)
        n2 = len(str2)

        result = ""

        j = 0
        i = 0
        while(i <= n1 - 1 and j <= n2 - 1):
            if (str1[i] != str2[j]):
                break
            result += (str1[i])

            i += 1
            j += 1

        return (result)

    def set_next_tokens(self, cur_input, tokens):
        add_EOI = False
        tokens = list(tokens)
        next_tokens = set()
        self.last_token = None
        self.case_insensitive = False

        for t in tokens:
            if t:
                if t.completable:
                    _, completions = t.complete(cur_input if cur_input else "")
                    if not completions:
                        next_tokens.add(t.helpstring)
                    else:
                        if t.case_insensitive:
                            self.case_insensitive = True
                        for c in completions:
                            next_tokens.add(str(c))
                else:
                    next_tokens.add(t.helpstring)
            else:
                add_EOI = True

        self.next_tokens = list(next_tokens)

        if add_EOI:
            self.next_tokens.append(EndOfInpuToken())
        elif cur_input:
            common_prefix = self.common_prefix(self.next_tokens)
            if common_prefix:
                self.last_token = (cur_input, common_prefix)

        if len(self.next_tokens) == 1 and self.next_tokens[0]:
            if self.last_token:
                if self.last_token[0] == self.last_token[1]:
                    if tokens[0].completable:
                        self.next_constant_token = self.last_token[1]
            else:
                if tokens[0].completable:
                    n, completion = tokens[0].complete("")
                    if completion and len(completion) == 1:
                        self.next_constant_token = str(completion[0])


class ExecContext():

    def __init__(self, interface, root_grammar, arglist):
        self._interface = interface
        self.print = interface.print
        self.error = interface.error
        self._root_grammar = root_grammar
        self._root_arglist = [NamedVariable(param.name) for param in root_grammar.param_list]
        self._param_mapping = map_grammar_arguments(root_grammar.name, root_grammar.param_list, arglist)

        for param in self._root_arglist:
            if param.var_id in self._param_mapping:
                param.assign(self._param_mapping[param.var_id].value)

        self._element_stack = []
        self._grammar_stack = []

    @property
    def root_arglist(self):
        return [e.value for e in self._root_arglist]

    def resolve_variable(self, var):
        if type(var) == DollarVariable:
            grammar_context = self._grammar_stack[-1] if self._grammar_stack else None
            if var in grammar_context.named_variables:
                return grammar_context.named_variables[var]
        elif type(var) == DollarNumber:
            parent_context = self._element_stack[-1] if self._element_stack else None
            if var in parent_context.token_variables:
                return parent_context.token_variables[var]
        return None

    async def resolve_argument(self, argument):
        if type(argument) in [DollarVariable, DollarNumber]:
            return self.resolve_variable(argument)
        elif type(argument) is CliArgument:
            return await self.resolve_argument(argument.value)
        elif type(argument) == BindingCall:
            ext_fn = argument.name
            arglist = [await self.evaluate(arg) for arg in argument.arglist]
            call_res = await self._interface.execute_binding_call(ext_fn, False, *arglist)
            return call_res
        elif type(argument) == FunctionCall:
            ext_fn = argument.name
            arglist = [await self.evaluate(arg) for arg in argument.arglist]
            call_res = await self._interface.execute_binding_call(ext_fn, True, *arglist)
            return call_res
        return argument

    async def evaluate(self, arg):
        res_arg = await self.resolve_argument(arg)
        if isinstance(res_arg, BindingVariable):
            res = res_arg.value
        else:
            res = res_arg

        if isinstance(res, str):
            res = convert_to_python_string(res)
        return res

    async def execute_binding(self, binding_code):

        grammar_context = self._grammar_stack[-1] if self._grammar_stack else None

        for code in binding_code:
            for block in code.blocks:
                if isinstance(block, AssignmentStatement):
                    lhs = None
                    if type(block.lhs) == DollarVariable:
                        lhs = self.resolve_variable(block.lhs)
                        if not lhs:
                            var = NamedVariable(str(block.lhs))
                            grammar_context.add_named_variable(var)
                            lhs = var
                    elif type(block.lhs) == DollarNumber:
                        lhs = self.resolve_variable(block.lhs)

                    if lhs:
                        rhs = await self.resolve_argument(block.rhs)
                        lhs.assign(rhs)

                elif isinstance(block, BindingCall):
                    ext_fn = block.name
                    arglist = [await self.evaluate(arg) for arg in block.arglist]
                    _ = await self._interface.execute_binding_call(ext_fn, False, *arglist)
                elif isinstance(block, FunctionCall):
                    ext_fn = block.name
                    arglist = [await self.evaluate(arg) for arg in block.arglist]
                    _ = await self._interface.execute_binding_call(ext_fn, True, *arglist)

    async def enter(self, element_ctx: TokenHierarchyElement, token_value: str):

        if element_ctx in self._element_stack:
            element_ctx = [e for e in self._element_stack if e == element_ctx][0]
            element_ctx.input_sequence.append(token_value)
            return

        element_ctx.input_sequence.append(token_value)

        if element_ctx not in self._element_stack:
            element = element_ctx.element

            if element.pre_match_binding:
                await self.execute_binding(element.pre_match_binding)

            if type(element) == NamedGrammar:
                if not self._element_stack:
                    for arg in self._root_arglist:
                        element_ctx.add_named_variable(arg)
                else:
                    grammar_ref_ctx = self._element_stack[-1]
                    for arg in grammar_ref_ctx.named_variables.values():
                         element_ctx.add_named_variable(arg)
                self._grammar_stack.append(element_ctx)

            elif type(element) == GrammarRefElement:
                arglist = [NamedVariable(param.name) for param in element.value.param_list]
                param_mapping = map_grammar_arguments(element.name, element.value.param_list, element.arg_list)
                for arg in arglist:
                    if arg.var_id in param_mapping:
                        res = await self.resolve_argument(param_mapping[arg.var_id])
                        arg.assign(res)
                        element_ctx.add_named_variable(arg)
            else:
                pass
            self._element_stack.append(element_ctx)

    async def exit(self, element_ctx: TokenHierarchyElement):
        if element_ctx not in self._element_stack:
            raise Exception("Context missing in stack. Recheck!!!")

        element_ctx = [e for e in self._element_stack if e == element_ctx][0]

        stack_ctx = self._element_stack.pop()
        if element_ctx != stack_ctx:
            raise Exception("Wrong context at top of stack. Recheck!!!")

        element = element_ctx.element
        parent_context = self._element_stack[-1] if self._element_stack else None
        grammar_context = self._grammar_stack[-1] if self._grammar_stack else None
        parent = parent_context.element if parent_context else None

        position = element.position
        if parent:
            if isinstance(parent, AlternativeInputElement):
                position = 0
            elif parent.repeat_count > 1:
                position = 0

        numbered_arg = TokenVariable("$" + str(position + 1))
        if len(element_ctx.input_sequence) == 1:
            numbered_arg.assign(element_ctx.input_sequence[0])
        else:
            numbered_arg.assign(element_ctx.input_sequence)

        if parent_context:
            parent_context.add_numbered_variable(numbered_arg)

        if element.post_match_binding:
            await self.execute_binding(element.post_match_binding)

        if type(element) == NamedGrammar:
            self._grammar_stack.pop()


class CliInterface(StdStreamsHolder):

    def __init__(self, loop, grammarset, stdin=None, stdout=None, stderr=None):

        if not isinstance(grammarset, GrammarSpecification):
            raise ValueError("GrammarSpecification object expected")

        self._loop = loop

        self.init_streams(stdin=stdin, stdout=stdout, stderr=stderr)

        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr

        self._tokens = {}
        self._grammars = grammarset
        self._grammar_stack = []
        self._token_class_map = None

    @property
    def loop(self):
        return self._loop

    @property
    def current_grammar(self):
        return self._grammar_stack[-1] if self._grammar_stack else None

    def enter_grammar(self, grammar_name):
        try:
            grammar = self._grammars.get_grammar(grammar_name)
            self._grammar_stack.append(grammar)
        except Exception as e:
            raise e

    def exit_grammar(self):
        try:
            self._grammar_stack.pop()
        except Exception as e:
            raise e

    def create_cli_token(self, token_name, tokendef=None):
        return CliToken(token_name)

    def get_token_classes(self):
        return []

    def get_token(self, name):
        if not (isinstance(name, str) and name):
            raise ValueError("Expected valid token name")

        if name not in self._tokens:
            tokendef = self._grammars.get_tokendef(name)
            if tokendef:
                if self._token_class_map is None:
                    token_classes = self.get_token_classes()
                    self._token_class_map = {t.__name__: t for t in token_classes}

                if tokendef.classname in self._token_class_map:
                    try:
                        self._tokens[name] = self._token_class_map[tokendef.classname](name, *tokendef.arglist)
                        return self._tokens[name]
                    except Exception as e:
                        self.error("Exception creating token object from token def:", e)

                _globals = globals()
                if tokendef.classname in _globals:
                    try:
                        self._tokens[name] =  _globals[tokendef.classname](name, *tokendef.arglist)
                        return self._tokens[name]
                    except Exception as e:
                        self.error("Exception creating token object from token def:", e)
            self._tokens[name] = self.create_cli_token(name, tokendef)

        return self._tokens[name]

    def get_cli_hook(self, func_name):
        return func_name

    async def resolve_local_function_call(self, func_name, *args, **kwarg):
        if func_name == 'list':
            l = []
            if args:
                l += args
            return l

        if func_name == 'print':
            return self.print(*args)

        if func_name == 'inc':
            r = args[0]
            try:
                r = r + 1
            except Exception:
                pass
            return r

        if func_name == 'add':
            r = args[0]
            for arg in args[1:]:
                try:
                    r = r + arg
                except Exception:
                    pass
            return r

        if func_name == 'dec':
            r = args[0]
            try:
                r = r + 1
            except Exception:
                pass
            return r

        if func_name == 'dict':
            return {}

        if func_name == 'set':
            s = set()
            for arg in args:
                s.add(arg)
            return s

        if func_name == 'append':
            if not args:
                raise ValueError("append called without arguments")
            r = args[0]
            for arg in args[1:]:
                if isinstance(r, list):
                    r += [arg]
                elif isinstance(r, set):
                    r.add(arg)
            return r

        if func_name == 'update':
            if len(args) > 2:
                r = args[0]
                if isinstance(r, dict):
                    r.update({args[1]: args[2]})
                return r
            return None

        if func_name == 'input':
            prompt = ""
            show_char = True
            if len(args):
                prompt = str(args[0])
                if len(args) > 1:
                    show_char = args[1]
            try:
                return await self.get_input(prompt, show_char)
            except:
                return ""

        return None

    async def execute_binding_call(self, func_name: str, local_function: bool, *args, **kwarg):
        try:
            ext_args = args

            if not local_function:
                if not hasattr(self, self.get_cli_hook(func_name)):
                    raise AttributeError("CLI interface missing function: {}".format(func_name))

                fn = getattr(self, self.get_cli_hook(func_name))

                if asyncio.iscoroutinefunction(fn):
                    res = await fn(*ext_args)
                else:
                    res = fn(*ext_args)
            else:
                res = await self.resolve_local_function_call(func_name, *ext_args, **kwarg)

            return res
        except Exception as e:
            self.error("Exception executing binding call:", type(e), e)

    async def execute_success_sequence(self, matched_sequence, match_values, arglist):

        exec_context = ExecContext(self, self.current_grammar, arglist)

        hierarchies = [token.get_element_hierarchy() for token in matched_sequence]
        token_values = match_values.copy()

        for _ in matched_sequence:
            hierarchy = hierarchies.pop(0)
            token_value = token_values.pop(0)

            h_copy = hierarchy.copy()
            while h_copy:
                parent = h_copy.pop()
                await exec_context.enter(parent, token_value)

            element = hierarchy.pop(0) if hierarchy else None

            await exec_context.exit(element)

            parent = hierarchy.pop(0) if hierarchy else None
            while(element and parent):
                if type(parent.element) in [AlternativeInputElement]:
                    await exec_context.exit(parent)
                elif len(parent.element) == (element.element.position + 1):
                    await exec_context.exit(parent)
                else:
                    rest_optional = True
                    position = element.element.position + 1
                    while position < len(parent.element):
                        next = parent.element.get(position)
                        if next.mandatory:
                            rest_optional = False
                            break
                        if hierarchies:
                            rest_optional = True
                            for h in hierarchies[0]:
                                if h.element == next:
                                    rest_optional = False
                                    break
                            if not rest_optional:
                                break
                        position += 1
                    if rest_optional:
                        await exec_context.exit(parent)
                    else:
                        break
                element = parent
                parent = hierarchy.pop(0) if hierarchy else None

        return exec_context.root_arglist

    async def match(self, tok_list, dry_run=False, last_token_complete=False, arglist=None):

        if not arglist:
            args = []
        else:
            args = []
            for arg in arglist:
                if isinstance(arg, CliArgument):
                    args.append(arg)
                else:
                    args.append(CliArgument(arg))

        cur_token_input = None
        token_list = tok_list.copy()

        prompt_choices = set(self.current_grammar.first(root_grammar=self.current_grammar))

        res = ParsingResult()

        if not prompt_choices:
            res.result = MATCH_FAILURE
            res.offending_token = None if not token_list else token_list[0]
            res.error = "No matching start tokens"
            return res

        def set_next_tokens(res, choices):
            next_tokens = set()
            add_EOT = False
            for c in choices:
                if not c:
                    add_EOT = True
                    continue
                next_tokens.add(self.get_token(c.name))
            if add_EOT:
                next_tokens.add(EndOfInpuToken())
            return res.set_next_tokens(None if last_token_complete else cur_token_input, next_tokens)

        initial = True
        seq_copy = []
        matching_sequences = []
        matching_seq_choices = []

        while True:
            seq_complete = False

            assert len(seq_copy) == len(matching_seq_choices)

            for sequence in seq_copy:

                full_matches = []
                partial_matches = []
                choices = matching_seq_choices.pop(0)

                for c in choices:
                    if self.get_token(c.name).match(cur_token_input) == MATCH_SUCCESS:
                        full_matches.append(sequence + [c])
                    if self.get_token(c.name).match(cur_token_input) == MATCH_PARTIAL:
                        partial_matches.append(sequence + [c])

                for f in full_matches:
                    matching_sequences.append(f)

                completion = []
                for p in partial_matches:
                    token = self.get_token(p[-1].name)
                    if token.completable:
                        n, comps = token.complete(cur_token_input)
                        completion += comps

                for p in partial_matches:
                    token = self.get_token(p[-1].name)
                    if token.completable:
                        if completion:
                            if dry_run and not token_list and not last_token_complete:
                                matching_sequences.append(p)
                            elif len(completion) == 1:
                                n, comps = token.complete(cur_token_input)
                                if len(comps) == 1:
                                    matching_sequences.append(p)
                            elif token.get_value(cur_token_input):
                                matching_sequences.append(p)
                            else:
                                res.result = MATCH_FAILURE
                                res.offending_token = tok_list[len(res.matched_sequence)]
                                res.offending_token_position = len(res.matched_sequence)
                                res.error = "Ambiguous options matched for the input token: {}".format(
                                    tok_list[len(res.matched_sequence)])
                                return res
                        elif dry_run and not token_list and not last_token_complete:
                            matching_sequences.append(p)
                        elif len(partial_matches) == 1:
                            v = token.get_value(cur_token_input)
                            if v is not None:
                                matching_sequences.append(p)
                    else:
                        v = token.get_value(cur_token_input)
                        if v is not None:
                            matching_sequences.append(p)
                        elif dry_run:
                            if not token_list and not last_token_complete:
                                matching_sequences.append(p)

            if not initial:

                if not matching_sequences:
                    if len(tok_list) == len(res.matched_sequence):
                        res.result = MATCH_PARTIAL
                        res.error = "Input sequence is not complete"
                    else:
                        res.result = MATCH_FAILURE
                        res.offending_token = cur_token_input
                        res.error = "Could not match any rule for this sequence"

                        if len(tok_list) > len(res.matched_sequence):
                            res.offending_token = tok_list[len(res.matched_sequence)]
                            res.offending_token_position = len(res.matched_sequence)
                    return res

                prompt_choices = set()

                for matching_sequence in matching_sequences:
                    choices = set()
                    last_token = matching_sequence[-1]
                    if last_token == EndOfInpuToken():
                        seq_complete = True
                    else:
                        if (not token_list) and (not last_token_complete):
                            choices.add(last_token)
                        else:
                            nexts = last_token.next()
                            for c in nexts:
                                if c == EndOfInpuToken():
                                    seq_complete = True
                                else:
                                    choices.add(c)
                    matching_seq_choices.append(choices)
                    prompt_choices = prompt_choices.union(choices)

            initial = False

            if not token_list:
                if seq_complete:
                    prompt_choices.add(EndOfInpuToken())

                if prompt_choices:
                    set_next_tokens(res, prompt_choices)
                else:
                    res.result = MATCH_FAILURE
                    if cur_token_input:
                        res.offending_token = cur_token_input
                    res.error = "Could not match any rule for this sequence"
                    return res

                if matching_sequences:
                    res.matched_sequence.append(cur_token_input)
                    if not dry_run:
                        if seq_complete and len(matching_sequences) > 1:
                            matching_sequences = self.fix_sequences(matching_sequences, tok_list)

                        if seq_complete:
                            if len(matching_sequences) == 1:
                                tok_index = 0
                                match_values = []
                                for t in matching_sequences[0]:
                                    token = self.get_token(t.name)
                                    match_value = token.get_value(tok_list[tok_index])
                                    match_values.append(match_value)
                                    tok_index += 1
                                res.matched_values = match_values
                                root_arglist = await self.execute_success_sequence(matching_sequences[0], match_values, args)
                                arglen = len(arglist)
                                for i in range(arglen):
                                    arglist[i] = root_arglist.pop(0)
                                res.result = MATCH_SUCCESS
                            else:
                                res.result = MATCH_AMBIGUOUS
                                res.error = "{} ambiguous sequences matched for the input".format(len(matching_sequences))
                        else:
                            if len(tok_list) == len(res.matched_sequence):
                                res.result = MATCH_PARTIAL
                                res.error = "Input sequence is not complete"
                            else:
                                res.result = MATCH_FAILURE
                                res.error = "Could not successfully match any rule with the input"
                        return res
                    else:
                        res.result = MATCH_PARTIAL
                        return res
            else:
                if cur_token_input:
                    res.matched_sequence.append(cur_token_input)
                cur_token_input = token_list.pop(0)
                if matching_sequences:
                    seq_copy = matching_sequences.copy()
                else:
                    seq_copy = [[]]
                    matching_seq_choices = [prompt_choices]
                matching_sequences = []

    def fix_sequences(self, matching_sequences, tok_list):
        if len(set(len(seq) for seq in matching_sequences)) == 1:
            seq_count = len(matching_sequences)
            seq_length = len(matching_sequences[0])

            tokens_to_keep = set([i for i in range(seq_count)])

            for i in range(seq_length):
                if len(set(seq[i] for seq in matching_sequences)) == 1:
                    continue
                else:
                    toc_types = []
                    for j in range(seq_count):
                        toc = matching_sequences[j][i]
                        token = self.get_token(toc.name)
                        toc_types.append(token.match(tok_list[i]))

                    if MATCH_SUCCESS in toc_types:
                        for j in range(seq_count):
                            if j in tokens_to_keep:
                                if toc_types[j] != MATCH_SUCCESS:
                                    tokens_to_keep.remove(j)

                    if len(tokens_to_keep) == 1:
                        return [matching_sequences[list(tokens_to_keep)[0]]]

                    toc_types = []
                    for j in range(seq_count):
                        toc = matching_sequences[j][i]
                        token = self.get_token(toc.name)
                        if token.completable:
                            toc_types.append(True)
                        else:
                            toc_types.append(False)

                    if True in toc_types:
                        for j in range(seq_count):
                            if j in tokens_to_keep:
                                if toc_types[j] != True:
                                    tokens_to_keep.remove(j)

                    if len(tokens_to_keep) == 1:
                        return [matching_sequences[list(tokens_to_keep)[0]]]
        else:
            return False, "Failed to resolve ambiguity of {} sequences with different lengths"
        return matching_sequences



# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import os
import sys
import shutil
from setuptools import setup


pkg_name = 'nessaid_cli'
test_pkg_name = 'nessaid_cli_tests'
sub_packages = ['binding_parser', 'tokenizer']

install_packages = [pkg_name, test_pkg_name] + [pkg_name + "." + sub_pkg for sub_pkg in sub_packages]

clanup_dirs = ['build', 'dist', pkg_name + '.egg-info']


def do_cleanup_fixes():
    dir_content = os.listdir()
    for d in clanup_dirs:
        if d in dir_content:
            try:
                shutil.rmtree(d)
            except Exception:
                pass


long_description = """Nessaid CLI Framework

This package implements a grammar specification and tools to compile the grammar and use it for driving CLIs.
The CLI commands can be expressed in grammar with the support for custom tokens with suggestion and autocompletion.

The grammar specification is parsed with PLY python package and the CLI objects are implemented on top of readline.
The command loop is in async mode.

Requirements

asyncio: The commandline interpreter will be looping for console input in async mode

ply: The lex-yacc like implementation of python. Will be used for parsing the grammar specification
and tokenizing the line input

pyreadline: For windows platforms. This is the readline implementation for windows platforms
Enduser Utilities

This package provides two classes for CLI building. NessaidCmd is intended to work
like the standard Cmd class for simple CLI designs.
It is a stripped down sub class of the more robest NessaidCli class.

NessaidCmd: The basic Cmd like tool for end user.
The CLI command handlers are defined as python methods with chosen prefix and the
grammar definitions to drive them are expressed as the method docstring.

NessaidCli: The generic base class for the CLI impelemntation.
NessaidCmd is the stripped down version as an alternative for the default Cmd implementation
"""

install_requires = [
    "asyncio",
    "ply",
]

if os.name == 'nt':
    install_requires.append("pyreadline")

setup(
    name=pkg_name,
    version='2.1.0',
    url='https://github.com/saithalavi/nessaid_cli',
    description="Nessaid's CLI tools",
    long_description=long_description,
    author='Saithalavi M',
    author_email='saithalavi@gmail.com',
    packages=install_packages,
    include_package_data=True,
    install_requires=install_requires,
    python_requires='>=3',
    keywords='cli parser cli-builder',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    project_urls = {
        'Documentation': 'https://github.com/saithalavi/nessaid_cli/blob/master/README.md',
        'Source': 'https://github.com/saithalavi/nessaid_cli',
        'Tracker': 'https://github.com/saithalavi/nessaid_cli/issues',
    },
    test_suite="nessaid_cli_tests",
)

if __name__ == '__main__':
    if 'clean' in sys.argv and 'install' not in sys.argv:
        do_cleanup_fixes()


# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

import unittest


class GrammarTest(unittest.TestCase):

    def test_dummy(self):
        pass

testcase1 = unittest.TestLoader().loadTestsFromTestCase(GrammarTest)

grammar_test = unittest.TestSuite([testcase1])


# Copyright 2021 by Saithalavi M, saithalavi@gmail.com
# All rights reserved.
# This file is part of the Nessaid CLI Framework, nessaid_cli python package
# and is released under the "MIT License Agreement". Please see the LICENSE
# file included as part of this package.
#

class BindingCode():

    def __init__(self, block_list):
        self._binding_blocks = block_list

    def __repr__(self):
        return "BindingCode({})".format(self._binding_blocks)

    def __str__(self):
        return self.__repr__()

    @property
    def blocks(self):
        return self._binding_blocks


class AssignmentStatement():

    def __init__(self, lhs, rhs):
        self._lhs = lhs
        self._rhs = rhs

    def __repr__(self):
        return "AssignmentStatement({} <- {})".format(self._lhs, self._rhs)

    def __str__(self):
        return self.__repr__()

    @property
    def lhs(self):
        return self._lhs

    @property
    def rhs(self):
        return self._rhs


class FunctionCall():

    def __init__(self, name, arglist):
        self._name = name
        self._arglist = arglist

    @property
    def name(self):
        return self._name

    @property
    def arglist(self):
        return self._arglist

    def __repr__(self):
        return "FunctionCall({}, {})".format(self._name, self._arglist)

    def __str__(self):
        return self.__repr__()

class BindingCall(FunctionCall):

    def __repr__(self):
        return "BindingCall({}, {})".format(self._name, self._arglist)

    def __str__(self):
        return self.__repr__()


class BindingStrObject(str):

    def __repr__(self):
        return '"' + self + '"'


class BindingVariable():

    def __init__(self):
        self._value = None

    def assign(self, value):
        if type(self._value) in [NamedVariable, TokenVariable]:
            self._value.assign(value)
        else:
            self._value = value

    @property
    def value(self):
        if type(self._value) in [NamedVariable, TokenVariable]:
            return self._value.value
        return self._value


class NamedVariable(BindingVariable):

    def __init__(self, dollar_name):
        self._name_id = dollar_name
        self._value = None

    def __repr__(self):
        return self._name_id

    def __str__(self):
        return self.__repr__()

    @property
    def var_id(self):
        return self._name_id

    @property
    def var_name(self):
        return self._name_id[1:]


class TokenVariable(BindingVariable):

    def __init__(self, dollar_num):
        self._num_id = dollar_num
        self._value = None

    @property
    def var_id(self):
        return self._num_id

    @property
    def var_index(self):
        return int(self._num_id[1:])

    def __repr__(self):
        return self._num_id

    def __str__(self):
        return self.__repr__()


